from __future__ import absolute_import, annotations

import functools as ft
import operator as op
import os
import platform
import re
from pathlib import Path
from typing import Any, Optional, Union

import attrs
import requests

from kslurm.appconfig import Config
from kslurm.args.command import CommandError
from kslurm.exceptions import ValidationError
from kslurm.models import validators
from kslurm.utils import get_hash

REGISTRYBASE = "https://registry-1.docker.io"
AUTHBASE = "https://auth.docker.io"
AUTHSERVICE = "registry.docker.io"


def find_(arg: str):
    container = SingularityDir().find(arg)
    if container is None:
        raise ValidationError(f"No image with identifier '{arg}' found")
    return container


@attrs.frozen
class URI:
    scheme: str
    org: str
    repo: str
    tag: str

    URI_SCHEME = "<scheme>://[<org>/]<image>[:<tag>]"

    @classmethod
    def parse(cls, uri: str):
        if not (
            match := re.match(
                r"""
                ^(?:(?P<scheme>[^:@\/]+)\:\/\/)?
                (?:(?P<org>[^:@\/]+)\/)?
                (?P<image>[^@:\/\s]+)
                (?::(?P<tag>[^@\/]+))?
                $
                """,
                uri,
                re.VERBOSE,
            )
        ):
            raise ValueError(f"Invalid uri. Should be {cls.URI_SCHEME}")

        uri_elems = match.groupdict()
        return cls(
            scheme=uri_elems["scheme"] or "",
            org=uri_elems["org"] or "",
            repo=uri_elems["image"] or "",
            tag=uri_elems["tag"] or "",
        )

    def __str__(self):
        return self.uri

    @property
    def uri(self):
        return "".join([f"{self.scheme}://" if self.scheme else "", self.address])

    @property
    def address(self):
        return "".join(
            [
                self.org + "/" if self.org else "",
                self.repo,
                ":" + self.tag if self.tag else "",
            ]
        )

    @property
    def image(self):
        return f"{self.org}/{self.repo}"


@attrs.frozen
class Container:
    uri: URI
    friendly_uri: Optional[URI] = None
    docker_data: Optional["DockerData"] = None

    @classmethod
    def from_uri(
        cls,
        uri: str,
        lookup_digest: bool = False,
        default: URI = URI(scheme="docker", repo="", org="library", tag=""),
    ):

        given = URI.parse(uri)
        if not given.tag:
            raise ValueError(f"Invalid uri. Must provide a tag. ({given.URI_SCHEME})")
        actual = attrs.evolve(
            default, **dict(filter(op.itemgetter(1), attrs.asdict(given).items()))
        )
        if lookup_digest and actual.scheme == "docker":
            data = DockerData.from_uri(actual)
        else:
            data = None
        return cls(uri=actual, friendly_uri=given, docker_data=data)

    @classmethod
    def from_uri_path(cls, path: Union[Path, str]):
        elems = Path(path).with_suffix("").parts
        if len(elems) > 4 or len(elems) < 3:
            raise ValueError(
                f"Incorrect number of parts in cache_path {path}. Expected 3 or 4"
            )
        if len(elems) == 4:
            tag = elems[3]
        else:
            tag = ""

        uri = URI(
            scheme=elems[0],
            org=elems[1],
            repo=elems[2],
            tag=tag or "",
        )
        friendly_uri = attrs.evolve(uri, org="") if uri.org == "library" else None
        return cls(uri=uri, friendly_uri=friendly_uri)

    @property
    def cache_path(self) -> Optional[str]:
        if self.docker_data and self.docker_data.trimmed_digest:
            return self.docker_data.trimmed_digest + ".sif"
        return None

    @property
    def uri_path(self):
        image_name = Path(self.uri.scheme, self.uri.org, self.uri.repo, self.uri.tag)
        return image_name.parent / (image_name.name + ".sif")

    @property
    def snakemake_cache_path(self):
        uri = self.friendly_uri or self.uri
        return Path(get_hash(uri.uri) + ".simg")

    def is_same_container(self, other: "Container"):
        return (
            other.uri.scheme == self.uri.scheme
            and other.uri.org == self.uri.org
            and other.uri.repo == self.uri.repo
        )

    def __str__(self):
        return str(self.friendly_uri or self.uri)


class ManifestError(CommandError):
    pass


@attrs.frozen
class DockerData:
    digest: str
    size: int

    @staticmethod
    def _get_target_arch():
        arch = platform.machine()
        if arch == "x86_64":
            return "amd64"
        elif arch.startswith("arm"):
            return "arm"
        elif arch.startswith("aarch64"):
            return "arm64"
        else:
            return None

    @staticmethod
    def _get_container_size(layers: list[dict[str, Any]]):
        return ft.reduce(op.add, map(op.itemgetter("size"), layers))

    @classmethod
    def from_uri(
        cls, uri: URI, digest: Optional[str] = None, token: Optional[str] = None
    ) -> Optional["DockerData"]:
        if token is None:
            token = requests.get(
                f"{AUTHBASE}/token",
                params=dict(
                    service=AUTHSERVICE,
                    scope=f"repository:{uri.image}:pull",
                ),
            ).json()["token"]

        digest = digest or uri.tag
        manifest_json = requests.get(
            f"{REGISTRYBASE}/v2/{uri.image}/manifests/{digest}",
            headers=dict(
                Authorization=f"Bearer {token}",
                Accept=(
                    "application/vnd.docker.distribution.manifest.v2+json, "
                    "application/vnd.docker.distribution.manifest.list.v2+json, "
                    "application/vnd.docker.distribution.manifest.v1+json"
                ),
            ),
        ).json()
        if "errors" in manifest_json:
            raise ManifestError(
                f"Unable to find '{uri}'. It may be a private repository, or you may "
                "have mispelled it."
            )

        schema_version = manifest_json["schemaVersion"]
        if schema_version == 2:
            media_type = manifest_json["mediaType"]
            if media_type == "application/vnd.docker.distribution.manifest.v2+json":
                return cls(
                    manifest_json["config"]["digest"],
                    size=cls._get_container_size(manifest_json["layers"]),
                )
            elif (
                media_type
                == "application/vnd.docker.distribution.manifest.list.v2+json"
            ):
                for manifest in manifest_json["manifests"]:
                    if manifest["platform"]["architecture"] == cls._get_target_arch():
                        return cls.from_uri(uri, digest=manifest["digest"], token=token)
        #     else:
        #         # fallback
        #         pass
        # elif schema_version == 1:
        #     # fallback
        #     pass
        # else:
        #     # fallback
        #     pass
        return None

    @property
    def trimmed_digest(self):
        if not (match := re.match(r".*\:(.*)", self.digest)):
            return None
        return match.group(1)

    @property
    def size_mb(self):
        return int(self.size / 1_000_000)


class SingularityDirError(CommandError):
    pass


class SingularityDir(type(Path())):
    def __new__(cls):
        pipdir = Config().get("pipdir")
        if pipdir is None:
            raise SingularityDirError(
                "pipdir not set. Please set pipdir using `kslurm config pipdir "
                "<directory>`, typically to a project-space or permanent storage "
                "directory"
            )
        pipdir = Path(pipdir)
        pipdir.mkdir(exist_ok=True)
        singularity_dir = pipdir / "containers"
        singularity_dir.mkdir(exist_ok=True)
        return super().__new__(cls, singularity_dir)

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.work.mkdir(exist_ok=True)
        self.images.mkdir(exist_ok=True)
        self.uris.mkdir(exist_ok=True)
        self.snakemake.mkdir(exist_ok=True)
        self.aliases.mkdir(exist_ok=True)

    @property
    def work(self):
        return self / "work"

    @property
    def images(self):
        return self / "images"

    @property
    def uris(self):
        return self / "uris"

    @property
    def snakemake(self):
        return self / "snakemake"

    @property
    def aliases(self):
        return self / "aliases"

    def get_data_path(self, container: Container):
        if container.cache_path:
            return self.images / container.cache_path
        result = self.uris / container.uri_path
        result.parent.mkdir(parents=True, exist_ok=True)
        return result

    def has_container(self, container: Container):
        if container.cache_path:
            return (self.images / container.cache_path).exists()
        return (self.uris / container.uri_path).exists()

    def has_raw_uri_file(self, container: Container):
        path = self.uris / container.uri_path
        return path.exists() and not path.is_symlink()

    def update_uri_link(self, container: Container):
        if container.cache_path:
            if (cache_path := self.images / container.cache_path).exists():
                uri_path = self.uris / container.uri_path
                if uri_path.exists():
                    if not uri_path.is_symlink():
                        os.remove(uri_path)
                    elif Path(cache_path).samefile(os.readlink(uri_path)):
                        return False
                    else:
                        os.unlink(uri_path)
                uri_path.parent.mkdir(exist_ok=True, parents=True)
                uri_path.symlink_to(cache_path)
                return True
            raise ValueError()
        return False

    def find(self, uri_or_alias: str):
        try:
            validators.fs_name(uri_or_alias)
        except ValidationError:
            container = None
        else:
            # Is it in the alias directory
            if (self.aliases / uri_or_alias).exists():
                container = ContainerAlias(uri_or_alias, self).image
            else:
                container = None

        if container is None:
            # Does it look like a uri
            try:
                container = Container.from_uri(uri_or_alias)
            except ValueError:
                raise SingularityDirError(
                    f"Invalid identifier '{uri_or_alias}'. Must be either an alias for "
                    "a container, or a valid container uri: "
                    f"'{URI.URI_SCHEME}'"
                )

        if (self.uris / container.uri_path).exists():
            return container
        return None

    def find_formatter(self, uri_or_alias: str):
        """Formatter version of find for use in argparser"""
        try:
            container = self.find(uri_or_alias)
        except SingularityDirError as err:
            raise ValidationError(err.msg)
        if container is None:
            raise ValidationError(f"No image with identifier '{uri_or_alias}' found")
        return container

    def iter_images(self):
        for dirpath, _, filenames in os.walk(self.uris):
            for filename in filenames:
                yield Path(dirpath, filename)


class AliasError(CommandError):
    pass


@attrs.define
class ContainerAlias:
    alias: str
    singularity_dir: SingularityDir
    _image: Optional[Container] = None

    def __attrs_post_init__(self):
        self.path.parent.mkdir(exist_ok=True)
        if self.path.exists():
            if not self.path.is_symlink():
                raise AliasError(
                    f"File or directory found at alias path '{self.path}'. Please "
                    "manually move or remove this item."
                )
            try:
                self.image
            except ValueError:
                raise AliasError(
                    f"Symlink found at '{self.path}' that does not currently point to "
                    "a kapp-controlled image. Please manually move or remove this item."
                )

    @property
    def path(self):
        return self.singularity_dir.aliases / self.alias

    @property
    def image(self):
        if self._image is None:
            self._image = Container.from_uri_path(
                Path(os.readlink(self.path)).relative_to(self.singularity_dir.uris)
            )
        return self._image

    def __bool__(self) -> bool:
        return self.path.exists()

    def check_upgrade(self, new_app: Container):
        if not self.image.is_same_container(new_app):
            raise AliasError(
                f"Alias '{self.alias}' is currently pointing to '{self.image.uri}'. To "
                "change this alias, run again with -f"
            )

    def link(self, app: Container):
        if self.path.is_symlink():
            if app == self.image:
                print(f"Alias '{self.alias}' already set to '{app}'")
                return
            print(f"Updating alias '{self.alias}' from {self.image} to {app}")
            self.path.unlink()
        else:
            print(f"Aliasing {app} as {self.alias}")
        self.path.symlink_to(self.singularity_dir.uris / app.uri_path)
        self._image = None


if __name__ == "__main__":
    container = Container.from_uri(
        "docker://akhanf/hippunfold:latest", lookup_digest=True
    )
    pass



from __future__ import absolute_import, annotations

from copy import copy
from typing import Any, Iterable, TypeVar, Union

from kslurm.args.arg import Arg, Context, Parser
from kslurm.exceptions import TailError

T = TypeVar("T", bound=Union[dict[str, Arg[Any]], type])
S = TypeVar("S")


def parse_args(
    args: Iterable[str],
    models: dict[str, Parser[Any]],
    allow_unknown: bool = True,
    terminate_on_unknown: bool = True,
):

    parsed, tail = _match_args(args, models, terminate_on_unknown)

    if tail and not allow_unknown:
        tail = TailError(f"{tail} does not match any args")

    return parsed, tail


def _match_args(
    args: Iterable[str],
    parsers: dict[str, Parser[Any]],
    terminate_on_unknown: bool = True,
):

    updated_params = copy(parsers)
    terminated = False

    tail: list[str] = []

    sorted_params = [
        parser.id
        for parser in sorted(
            parsers.values(), key=lambda param: param.priority, reverse=True
        )
    ]
    arg_list = list(args)
    last_match = None
    for i, arg in enumerate(arg_list):
        if terminated:
            tail.append(arg)
            continue
        next_fragment = arg
        while isinstance(next_fragment, str):
            fragment: str = next_fragment

            context = Context(
                args=arg_list,
                current_arg=i,
                params=updated_params,
                last_matched=last_match,
            )
            matched_arg = False
            for id in sorted_params:
                parser = updated_params[id]
                if next_fragment := parser.match(fragment, parser, context):
                    updated_params[id] = parser.with_value(fragment, context)
                    if parser.terminal:
                        terminated = True
                    last_match = updated_params[parser.id]
                    matched_arg = True
                    break
            if not matched_arg:
                next_fragment = False
                tail.append(arg)
                if terminate_on_unknown:
                    terminated = True

    return updated_params, tail



from __future__ import absolute_import

import os

import attrs

from kslurm.args import CommandError, Subcommand, command, flag, positional, subcommand
from kslurm.container import Container, SingularityDir


@command
def _list():
    """List all available containers"""
    singularity_dir = SingularityDir()
    for path in singularity_dir.iter_images():
        print(Container.from_uri_path(path.relative_to(singularity_dir.uris)))


@command(inline=True)
def _rm(
    uri_or_alias: str = positional(),
    purge: bool = flag(
        ["-p", "--purge"], help="Remove all underlying data files (not yet implemented)"
    ),
):
    """Remove a docker image uri.

    By default, only the uri pointer is removed; underlying datafiles are kept. In a
    future release, the purge flag will be available for deleting the underlying data.

    To remove datafiles, run `kapp purge dangling`
    """
    singularity_dir = SingularityDir()
    container = singularity_dir.find(uri_or_alias)
    if container is None:
        raise CommandError(f"No image with identifier '{uri_or_alias}' found")

    path = singularity_dir.uris / container.uri_path

    if not path.is_symlink() and not purge:
        raise CommandError(
            f"The uri {container.uri}' is directly attached to it's data and cannot "
            "be removed without the --purge flag."
        )

    if path.is_symlink() and purge:
        print("[INFO] The purge flag is not yet implemented in this context.")

    for snakemake_alias in singularity_dir.snakemake.iterdir():
        if snakemake_alias.is_symlink() and path.samefile(snakemake_alias):
            snakemake_alias.unlink()
            if path.is_symlink():
                snakemake_alias.symlink_to(os.readlink(path))

    for alias in singularity_dir.aliases.iterdir():
        if alias.is_symlink() and path.samefile(alias):
            alias.unlink()
            if uri_or_alias != alias.name:
                print(f"Removing alias '{alias.name}'")

    if path.is_symlink():
        path.unlink()
    else:
        os.remove(path)


@attrs.frozen
class _ImgListModel:
    command: Subcommand = subcommand(
        commands={
            "list": _list.cli,
            "rm": _rm.cli,
        },
    )


@command
def img_cmd(cmd_name: str, args: _ImgListModel, tail: list[str]):
    """Interact with container image files"""
    name, func = args.command
    entry = f"{cmd_name} {name}"
    return func([entry, *tail])



from __future__ import absolute_import

import importlib.resources as impr
from pathlib import Path

import attr

import neuroglia_helpers
from kslurm.args import Subcommand, command, error, flag, subcommand
from kslurm.cli.config import config
from kslurm.cli.kbatch import kbatch
from kslurm.cli.kjupyter import kjupyter
from kslurm.cli.kpy import kpy
from kslurm.cli.krun import krun
from kslurm.installer.installer import install

NAME = "kslurm"
HOME_DIR = "KSLURM_HOME"
ENTRYPOINTS = ["kbatch", "krun", "kjupyter", "kslurm", "kpy"]


@command(inline=True)
def _neuroglia_helpers(show_src: bool = flag(["--src-dir"])):
    if show_src:
        print(Path(neuroglia_helpers.__file__).parent)
    else:
        with impr.path("kslurm.bin", "neuroglia-helpers.sh") as path:
            print(f"\nsource {path.resolve()}")


@attr.s(auto_attribs=True)
class KslurmModel:
    command: Subcommand = subcommand(
        commands={
            "kbatch": kbatch,
            "krun": krun,
            "kjupyter": kjupyter,
            "kpy": kpy,
            "config": config,
            "update": error,
            "neuroglia-helpers": _neuroglia_helpers,
        },
    )


@command
def main(args: KslurmModel, tail: list[str]) -> int:
    name, func = args.command
    entry = f"kslurm {name}"
    if name == "update":
        install(tail, NAME, HOME_DIR, ENTRYPOINTS)
    return func([entry, *tail])


if __name__ == "__main__":
    main.cli(["kslurm"])



from __future__ import absolute_import

import importlib.resources as impr
import os
import re
import shutil
import subprocess as sp
import sys
import tarfile
import tempfile
from pathlib import Path
from typing import Literal, Optional, overload

import attr
from shellingham import ShellDetectionFailure

from kslurm.appconfig import InvalidPipdirError
from kslurm.args import Subcommand, choice, flag, keyword, positional, shape, subcommand
from kslurm.args.command import CommandError, command
from kslurm.args.help import SKIPHELP
from kslurm.models import validators
from kslurm.shell import Shell
from kslurm.venv import KpyIndex, PromptRefreshError, VenvCache, VenvPrompt, rebase_venv


def _get_unique_name(index: KpyIndex, stem: str = "venv", i: int = 0) -> str:
    if i == 0:
        candidate = stem
    else:
        candidate = f"{stem}{i}"
    if candidate in index:
        return _get_unique_name(index, stem, i + 1)
    return candidate


def _get_shell():
    try:
        return Shell.get()
    except ShellDetectionFailure as err:
        raise CommandError(err.args[0])


class MissingSlurmTmpdirError(CommandError):
    pass


@overload
def _get_slurm_tmpdir(allow_missing: Literal[True] = ...) -> Optional[Path]:
    ...


@overload
def _get_slurm_tmpdir(allow_missing: Literal[False] = ...) -> Path:
    ...


def _get_slurm_tmpdir(allow_missing: bool = True):
    if not os.environ.get("SLURM_TMPDIR"):
        if not allow_missing:
            raise MissingSlurmTmpdirError(
                "This command can only be used in a compute node. Use `krun` to start "
                "an interactive session"
            )
        return
    return Path(os.environ["SLURM_TMPDIR"])


@command
def _bash():
    """Echo script for inclusion in .bashrc

    e.g.
        kpy bash >> $HOME/.bashrc
    """
    with impr.path("kslurm.bin", "bash.sh") as path:
        print(f"\nsource {path.resolve()}")


@command(inline=True)
def _load(
    name: str = positional(default=""),
    new_name: str = keyword(match=["--as"], format=validators.fs_name),
    script: str = keyword(match=["--script"], help=SKIPHELP),
):
    """Load a saved python venv

    Run without name to list available venvs for loading

    Attributes:
        name: Name of the venv to load. If not specified, list all available venvs
        new_name:
            Load the venv under a different name. Useful for loading the same venv twice
    """
    slurm_tmp = _get_slurm_tmpdir()
    if slurm_tmp:
        index = KpyIndex(slurm_tmp)

        label = new_name or name
        if label in index:
            raise CommandError(
                f"An environment called '{label}' already exists. You can load "
                f"'{name}' under a different name using --as:\n"
                f"\tkpy load {label} --as <name>\n"
                f"You can also activate the existing '{label}' using\n"
                f"\tkpy activate {label}"
            )
        venv_dir = Path(tempfile.mkdtemp(prefix="kslurm-venv-", dir=slurm_tmp / "tmp"))
    else:
        index = None
        label = name
        venv_dir = Path(tempfile.mkdtemp(prefix="kslurm-"))

    venv_cache = VenvCache()

    if not name or name not in venv_cache:
        print("Valid venvs:\n" + str(venv_cache))
        return

    print(f"Unpacking venv '{name}'", end="")
    if label != name:
        print(f" as '{label}'")
    else:
        print()
    with tarfile.open(venv_cache[name], "r") as tar:
        tar.extractall(venv_dir)

    print("Updating paths")
    rebase_venv(venv_dir)

    prompt = VenvPrompt(venv_dir)
    prompt.update_prompt(label)
    prompt.update_hash()
    prompt.save()

    if index is not None:
        index[label] = str(venv_dir)
        index.write()

    shell = _get_shell()
    if script:
        with Path(script).open("w") as f:
            f.write(shell.source(venv_dir))
        return 2
    shell.activate(venv_dir)


@command(inline=True)
def _export(
    mode: str = choice(["venv"], help="What sort of export to perform"),
    name: str = positional(),
    path: Path = keyword(["--path", "-p"], default=None),
):
    """Export a saved venv

    Saves to a path of choice. Currently "venv" is the only valid export mode. Exported
    venvs can only be safely activated by a bash shell.

    Attributes:
        name: Name of the venv to export
        path: Path for the export
    """
    venv_cache = VenvCache()

    if name not in venv_cache:
        raise CommandError("Valid venvs:\n" + str(venv_cache))

    if path.exists():
        raise CommandError(f"{path} already exists")

    print("exporting...")
    with tarfile.open(venv_cache[name], "r") as tar:
        tar.extractall(path)
    rebase_venv(path)

    print(
        "Export complete! Activate the venv by running\n\tsource "
        f"{path}/bin/activate"
    )


@command(inline=True)
def _save(
    name: str = positional(format=validators.fs_name),
    force: bool = flag(match=["--force", "-f"]),
):
    """Save current venv

    Attributes:
        name:
            Name to save venv under. If a venv of this name already exists, an error
            will be thrown, unless force is used
        force:
            Overwrite any existing venv with chosen name
    """
    if not os.environ.get("VIRTUAL_ENV"):
        raise CommandError(
            "No active virtual env detected. Please activate one, or ensure "
            "$VIRTUAL_ENV is being set correctly"
        )
    venv_cache = VenvCache()

    delete = False
    if name in venv_cache:
        if force:
            delete = True
        else:
            print(f"{name} already exists. Run with -f to force overwrite")
            return

    dest = venv_cache.get_path(name)

    _, tmp = tempfile.mkstemp(prefix="kslurm-", suffix=".tar.gz")

    venv_dir = Path(os.environ["VIRTUAL_ENV"])
    prompt = VenvPrompt(venv_dir)
    prompt.update_prompt(name)
    prompt.update_hash()
    prompt.save()
    with tarfile.open(tmp, mode="w:gz") as tar:
        tar.add(venv_dir, arcname="")

    slurm_tmp = _get_slurm_tmpdir()
    if slurm_tmp:
        index = KpyIndex(slurm_tmp)
        index[name] = str(venv_dir)
        index.write()

    # Do a two stage move in case tmp and dest are on different file systems, which
    # could make the move take some time. This lets us delete the old dest at the last
    # possible second
    stage = dest.with_suffix(".tar.gz.tmp")
    shutil.move(tmp, stage)
    if delete:
        os.remove(dest)
    shutil.move(stage, dest)


@command(inline=True)
def _create(
    name: Optional[str] = positional(format=validators.fs_name),
    version: Optional[str] = shape(
        match=r"^[23]\.\d{1,2}$",
        examples=["2.7", "3.8"],
    ),
    script: str = keyword(match=["--script"], help=SKIPHELP),
):
    """Create a new venv

    If no name provided, a placeholder name will be generated

    Attributes:
        name: Name of the new venv

        version:
            Python version to use in new venv. An appropriate executable must be on the
            $PATH (e.g. 3.7 -> python3.7)

            @help.syntax (2|3).x
    """
    if version:
        ver = ["-p", version]
    else:
        try:
            data = sp.run(
                "eval $($LMOD_CMD bash list python)", shell=True, capture_output=True
            )
            if match := re.search(r"(?<=python\/)\d\.\d{1,2}", data.stdout.decode()):
                ver = ["-p", match[0]]
            else:
                ver = []
        except RuntimeError:
            ver = []

    slurm_tmp = _get_slurm_tmpdir()
    if slurm_tmp:
        index = KpyIndex(slurm_tmp)
        name = name if name else _get_unique_name(index, "venv")
        if name in index:
            raise CommandError(
                f"An environment called '{name}' already exists. You can activate "
                f"the existing '{name}' using\n"
                "\tkpy activate {name}"
            )

        venv_dir = tempfile.mkdtemp(prefix="kslurm-venv-", dir=slurm_tmp / "tmp")
        no_download = ["--no-download"]
        no_index = ["--no-index"]
    else:
        index = None
        name = name if name else "venv"
        venv_dir = tempfile.mkdtemp(prefix="kslurm-")
        no_download = []
        no_index = []

    try:
        sp.run(
            [
                sys.executable,
                "-m",
                "virtualenv",
                venv_dir,
                "--symlinks",
                *ver,
                *no_download,
            ],
        )
        sp.run(
            [
                os.path.join(venv_dir, "bin", "python"),
                "-m",
                "pip",
                "install",
                "--upgrade",
                "pip",
                *no_index,
            ],
        )
    except RuntimeError as err:
        print(err.args[0])
        return 1
    if index is not None:
        index[name] = str(venv_dir)
        index.write()
    prompt = VenvPrompt(Path(venv_dir))
    prompt.update_prompt(name)
    prompt.save()

    shell = _get_shell()
    if script:
        with Path(script).open("w") as f:
            f.write(shell.source(Path(venv_dir)))
        return 2
    shell.activate(Path(venv_dir))


@command(inline=True)
def _activate(
    name: Optional[str] = positional(),
    script: str = keyword(["--script"], help=SKIPHELP),
):
    """Activate a venv already created or loaded

    Only works on compute nodes. Use kpy create or kpy load --as on a login node

    Attributes:
        name: Name of the venv to activate. If not provided, list all loaded venvs
    """
    slurm_tmp = _get_slurm_tmpdir(False)
    index = KpyIndex(slurm_tmp)
    if not name:
        print(str(index))
        return

    if name not in index:
        err = [f"An environment with the name '{name}' has not yet been initialized. "]
        try:
            venv_cache = VenvCache()
            if name in venv_cache:
                err += [
                    f"The saved environment called '{name}' can be loaded using",
                    f"\tkpy load {name}\n",
                ]
        except InvalidPipdirError:
            pass
        err += [
            f"A new environment can be created using\n\tkpy create {name}\n",
            f"Currently initialized environments:\n{index}",
        ]
        raise CommandError("\n".join(err))

    shell = _get_shell()
    if script:
        with Path(script).open("w") as f:
            f.write(shell.source(Path(index[name])))
        return 2
    shell.activate(Path(index[name]))


@command
def _list():
    """List all saved venvs.

    To list initialized venvs (either created or loaded), run `kpy activate` without any
    arguments
    """
    venv_cache = VenvCache()
    print(str(venv_cache))
    return


@command
def _refresh():
    try:
        dir = Path(os.environ["VIRTUAL_ENV"])
    except KeyError:
        return
    prompt = VenvPrompt(dir)
    try:
        prompt.refresh()
        print(prompt.name)
    except PromptRefreshError:
        return


def _kpy_wrapper(argv: list[str] = sys.argv):
    with impr.path("kslurm.bin", "kpy-wrapper.sh") as path:
        print(path)
    return 0


@command(inline=True)
def _rm(name: Optional[str] = positional()):
    """Delete a venv
    Args:
        name (Optional[str], optional):
            Name of the venv to delete

    """
    venv_cache = VenvCache()

    if not name:
        raise CommandError("Valid venvs:\n" + str(venv_cache))

    if name not in venv_cache:
        raise CommandError(
            f"{name} is not a valid venv. Currently saved venvs are:\n{venv_cache}"
        )

    os.remove(venv_cache[name])


@attr.frozen
class _KpyModel:
    command: Subcommand = subcommand(
        commands={
            "load": _load,
            "save": _save,
            "bash": _bash,
            "create": _create,
            "activate": _activate,
            "list": _list,
            "rm": _rm,
            "export": _export,
            "_refresh": _refresh,
            "_kpy_wrapper": _kpy_wrapper,
        },
    )


@command
def kpy(cmd_name: str, args: _KpyModel, tail: list[str]):
    """Set of commands for interacting with python virtual envs"""
    name, func = args.command
    entry = f"{cmd_name} {name}"
    return func([entry, *tail])


if __name__ == "__main__":
    kpy.cli(["kpy", "foo"])



from __future__ import absolute_import

import os

import attrs

from kslurm.args import CommandError, Subcommand, command, positional, subcommand
from kslurm.container import SingularityDir


@command
def _list():
    """List all available aliases"""
    singularity_dir = SingularityDir()
    for alias in singularity_dir.aliases.iterdir():
        container = singularity_dir.find(alias.name) or "-INVALID-"
        print(f"{alias.name} -> {container}")


@command(inline=True)
def _rm(
    alias: str = positional(),
):
    """Remove an alias"""
    singularity_dir = SingularityDir()
    path = singularity_dir.aliases / alias
    if not path.exists():
        raise CommandError(f"'{alias}' is not a valid alias")
        return 1
    if path.is_symlink():
        path.unlink()
    elif path.exists():
        os.remove(path)


@attrs.frozen
class _AliasModel:
    command: Subcommand = subcommand(
        commands={
            "list": _list.cli,
            "rm": _rm.cli,
        },
    )


@command
def alias_cmd(cmd_name: str, args: _AliasModel, tail: list[str]):
    """Interact with container image files"""
    name, func = args.command
    entry = f"{cmd_name} {name}"
    return func([entry, *tail])



# # pyright: reportPrivateUsage=false
# from __future__ import absolute_import

from __future__ import absolute_import

# from pathlib import Path
# from pytest_mock import MockerFixture

# import kslurm.installer.installer as installer


# class TestMakeEnv:
#     def test_creates_new_venv(self, mocker: MockerFixture, tmp_path: Path):
#         envbuilder = mocker.patch.object(installer, "EnvBuilder")
#         assert installer._make_env(tmp_path) == tmp_path / "venv"
#         envbuilder.assert_called_once()

#     def test_doesnt_create_venv_if_update_and_venv_exists(
#         self, mocker: MockerFixture, tmp_path: Path
#     ):
#         envbuilder = mocker.patch.object(installer, "EnvBuilder")
#         mocker.patch.object(installer, "_is_venv", return_value=True)
#         assert installer._make_env(tmp_path, True) == tmp_path / "venv"
#         envbuilder.assert_not_called()

#     def test_creates_venv_if_venv_exists_but_update_not_true(
#         self, mocker: MockerFixture, tmp_path: Path
#     ):
#         envbuilder = mocker.patch.object(installer, "EnvBuilder")
#         mocker.patch.object(installer, "_is_venv", return_value=True)
#         assert installer._make_env(tmp_path) == tmp_path / "venv"
#         envbuilder.assert_called_once()

#     def test_creates_venv_if_update_but_no_existing_venv(
#         self, mocker: MockerFixture, tmp_path: Path
#     ):
#         envbuilder = mocker.patch.object(installer, "EnvBuilder")
#         assert installer._make_env(tmp_path, True) == tmp_path / "venv"
#         envbuilder.assert_called_once()


# def test_install_library(mocker: MockerFixture, tmp_path: Path):
#     sub = mocker.patch("subprocess.run")
#     installer._install_library("pkg==3.2", tmp_path)
#     sub.assert_called_once()



from __future__ import absolute_import

import os
import os.path
import subprocess as sp
import tarfile
from pathlib import Path
import shutil
from typing import DefaultDict, Optional

import attrs
import requests
from InquirerPy import inquirer as inq  # type: ignore

from kslurm.appcache import Cache
from kslurm.args import (
    CommandError,
    Subcommand,
    choice,
    command,
    flag,
    keyword,
    positional,
    subcommand,
)
from kslurm.cli.kapp.alias import alias_cmd
from kslurm.cli.kapp.image import img_cmd
from kslurm.cli.krun import krun
from kslurm.container import Container, ContainerAlias, SingularityDir
from kslurm.models import formatters, validators
from kslurm.utils import get_hash

_SINGULARITY_DIR = SingularityDir()


def _update_aliases(
    singularity_dir: SingularityDir,
    app: Container,
    uri: str,
    alias: Optional[ContainerAlias],
):
    if singularity_dir.update_uri_link(app) and app.docker_data:
        print(f"Updated '{uri}' to '{app.docker_data.digest}'")
    else:
        print(f"{uri} already up to date")
    if alias is not None:
        alias.link(app)

    if (
        snakemake_path := singularity_dir.snakemake / app.snakemake_cache_path
    ).is_symlink():
        snakemake_path.unlink()
    elif snakemake_path.exists():
        os.remove(snakemake_path)
    snakemake_path.symlink_to(singularity_dir.get_data_path(app))


def _check_singularity():
    if sp.run(["command", "-v", "singularity"], capture_output=True).returncode != 0:
        raise CommandError(
            "singularity command is not available. Try running `module load singularity"
        )


@command(inline=True)
def _pull(
    uri: str = positional(),
    alias_name: str = keyword(
        ["-a", "--alias"],
        help="Alias for the docker container",
        format=validators.fs_name,
    ),
    force: bool = flag(["-f", "--force"], help="Force overwriting of aliases"),
    mem: int = keyword(
        ["--mem"],
        format=formatters.mem,
        help="Amount of memory to allocate when building image",
    )
    # executable: bool = flag(
    #     ["-x", "--exe"], help="Include --name as an executable on the $PATH"
    # ),
):
    """Pull a container. Defaults to docker hub."""
    _check_singularity()

    try:
        app = Container.from_uri(uri, lookup_digest=True)
    except ValueError as err:
        raise CommandError(err.args[0])

    if app.uri.scheme != "docker":
        raise CommandError(
            f"Invalid scheme '{app.uri.scheme}'. Only 'docker://' uris supported"
        )

    if alias_name:
        alias = ContainerAlias(alias_name, _SINGULARITY_DIR)
        if alias and not force:
            alias.check_upgrade(app)
    else:
        alias = None

    if _SINGULARITY_DIR.has_container(app):
        _update_aliases(_SINGULARITY_DIR, app, uri, alias)
        return

    if _SINGULARITY_DIR.has_raw_uri_file(app):
        if not inq.confirm(
            f"An image matching {app.uri.uri} already exists, but we can't verify if "
            "it's up to date. Would you like to pull it again?"
        ).execute():
            return

    image_path = _SINGULARITY_DIR.get_data_path(app)

    # Small images we can directly use the singularity command
    if app.docker_data and app.docker_data.size_mb < 200 and not mem:
        sp.run(["singularity", "pull", str(image_path), app.uri.uri])
        _update_aliases(_SINGULARITY_DIR, app, uri, alias)
        return

    url = (
        "https://raw.githubusercontent.com/moby/moby/v20.10.16/contrib/"
        "download-frozen-image-v2.sh"
    )
    cache = Cache()
    try:
        script = cache[url]
    except KeyError:
        r = requests.get(
            "https://raw.githubusercontent.com/moby/moby/master/contrib/"
            "download-frozen-image-v2.sh",
            headers={"user-agent": "kslurm"},
        )

        if r.status_code >= 400:
            raise CommandError("Unable to retrieve container download script")
        script = r.text
        # hashed = impr.read_text("kslurm.data", "download_frozen_container.hash")
        # if get_hash(script, method="sha512") != hashed.strip():
        #     raise CommandError(
        #         "Frozen image download script does not match hash. Cannot continue."
        #     )
        cache[url] = script
        os.chmod(cache.get_path(url), 0o776)

    workdir = _SINGULARITY_DIR.work / get_hash(app.uri.address)
    frozen_image = workdir / "image"
    proc = sp.run(
        [
            str(cache.get_path(url)),
            str(frozen_image),
            (app.friendly_uri or app.uri).address,
        ]
    )
    if proc.returncode != 0:
        return 1
    with tarfile.open(frozen_image.with_suffix(".tar"), "w") as tar:
        for path in frozen_image.iterdir():
            tar.add(path, arcname=path.relative_to(frozen_image))

    # Check that image_path is not a broken symlink
    if not image_path.exists() and image_path.is_symlink():
        image_path.unlink()
    mem = (
        mem
        if mem
        else (min(64000, app.docker_data.size_mb * 20) if app.docker_data else 8000)
    )
    ret = krun.cli(
        [
            "krun",
            "1:00",
            f"{mem}M",
            "singularity",
            "build",
            "--disable-cache",
            str(image_path),
            f"docker-archive://{frozen_image.with_suffix('.tar')}",
        ]
    )
    if ret == 0:
        shutil.rmtree(workdir)

        _update_aliases(_SINGULARITY_DIR, app, uri, alias)
    return ret


@command(inline=True)
def _path(
    uri_or_alias: str = positional(),
    quiet: bool = flag(["-q", "--quiet"], help="Don't print any errors"),
):
    """Print the path of the given uri or alias"""
    try:
        container = _SINGULARITY_DIR.find(uri_or_alias)
    except Exception as err:
        if not quiet:
            raise err
        return 1

    if container is None:
        raise CommandError(f"No image with identifier '{uri_or_alias}' found")
    print(_SINGULARITY_DIR.get_data_path(container))


@attrs.frozen
class _RunModel:
    container: Container = positional(
        format=_SINGULARITY_DIR.find_formatter, name="uri_or_alias"
    )


def _generic_run(container: Container, cmd: str, args: list[str]):
    _check_singularity()
    sp.run(
        [
            "singularity",
            cmd,
            _SINGULARITY_DIR.get_data_path(container),
            *args,
        ]
    )


@command
def _run(args: _RunModel, container_args: list[str]):
    _generic_run(args.container, "run", container_args)


@command
def _shell(args: _RunModel, container_args: list[str]):
    _generic_run(args.container, "shell", container_args)


@command
def _exec(args: _RunModel, container_args: list[str]):
    _generic_run(args.container, "exec", container_args)


@command(inline=True)
def _purge(
    scope: str = choice(["dangling"]),
    dry: bool = flag(
        ["-d", "--dry"],
        help="Print what files will be deleted without deleting anything",
    ),
):
    """Remove dangling containers

    Remove all containers not referred to by any uris (e.g. because the uri was removed)
    """
    uri_list = set(
        os.path.basename(os.readlink(path))
        for path in _SINGULARITY_DIR.iter_images()
        if path.is_symlink()
    )
    snakemake_aliases: dict[str, list[Path]] = DefaultDict(list)
    for path in _SINGULARITY_DIR.snakemake.iterdir():
        if path.is_symlink():
            snakemake_aliases[os.path.basename(os.readlink(path))].append(path)
    count = 0
    for file in _SINGULARITY_DIR.images.iterdir():
        if file.name in uri_list:
            continue
        if dry:
            print(file.name)
        else:
            for path in snakemake_aliases[file.name]:
                path.unlink()
            os.remove(file)
            count += 1

    if not dry:
        if not count:
            print("Nothing to remove")
        elif count == 1:
            print("Removed 1 file")
        else:
            print(f"Removed {count} files")


@command
def _snakemake():
    """Print the snakemake singularity directory (to be used with --singularity-prefix)

    Snakemake supports supplying a --singularity prefix: a directory will image files
    will be searched and saved. By supplying the path printed by this command, snakemake
    will automatically use any containers pulled using kapp
    """
    print(_SINGULARITY_DIR.snakemake.resolve())


@attrs.frozen
class _KappModel:
    command: Subcommand = subcommand(
        commands={
            "pull": _pull.cli,
            "path": _path.cli,
            "image": img_cmd.cli,
            "run": _run.cli,
            "shell": _shell.cli,
            "exec": _exec.cli,
            "alias": alias_cmd.cli,
            "purge": _purge.cli,
            "snakemake": _snakemake.cli,
        },
    )


@command
def kapp(cmd_name: str, args: _KappModel, tail: list[str]):
    """Set of commands for interacting with containers"""
    name, func = args.command
    entry = f"{cmd_name} {name}"
    return func([entry, *tail])



from __future__ import absolute_import

import re
import subprocess as sp
from typing import Any, DefaultDict

from InquirerPy import inquirer as inq  # type: ignore
from InquirerPy.base import Choice  # type: ignore

import kslurm.appconfig as appconfig
from kslurm.args import flag, positional
from kslurm.args.command import CommandError, command


class InteractiveConfigError(CommandError):
    pass


def _get_account_type(
    accounts: dict[str, set[str]], levelfs: dict[str, str], label: str
) -> str:
    choices: list[Any] = [
        Choice(account, f"{account} (LevelFS: {levelfs[account + '_' + label]})")
        for account, types in accounts.items()
        if label in types
    ]
    if not choices:
        print(f"No {label}-enabled accounts available. Skipping.")
        return ""
    elif len(choices) == 1:
        print(f"One {label}-enabled account available: {choices[0]}")
        return choices[0].value
    else:
        return inq.fuzzy(
            f"Select an account for {label} jobs. Higher LevelFS is better.",
            choices=choices,
        ).execute()


def _account():
    whoami = sp.run(["whoami"], capture_output=True).stdout.decode().strip()
    proc = sp.run(
        [
            "sacctmgr",
            "list",
            "user",
            "WithAssoc",
            f"Names={whoami}",
            "Format=Account",
            "-Pn",
        ],
        capture_output=True,
    )
    if proc.returncode != 0:
        raise InteractiveConfigError(
            "sacctmgr command not available. Cannot interactively set account."
        )
    accounts = set(proc.stdout.decode().strip().splitlines())
    account_types: dict[str, set[str]] = DefaultDict(set)
    account_fs: dict[str, str] = {}
    for account in accounts:
        if not (match := re.match(r"^([\w\-]+)_(cpu|gpu)$", account)):
            continue
        account_fs[account] = (
            sp.run(
                ["sshare", "-A", account, "-o", "LevelFS", "-hP"], capture_output=True
            )
            .stdout.decode()
            .strip()
            .splitlines()[0]
        )
        account_types[match[1]].add(match[2])
    if not account_types:
        raise InteractiveConfigError(
            "No accounts found using sacctmgr. Cannot interactively set account."
        )
    cpu_choice = _get_account_type(account_types, account_fs, "cpu")
    gpu_choice = _get_account_type(account_types, account_fs, "gpu")
    return {"account.gpu": gpu_choice, "account.cpu": cpu_choice, "account": None}


INTERACTIVE_SETTINGS = {"account": _account}


@command(inline=True)
def config(
    entry: str = positional(help="Configuration value to update"),
    value: str = positional(
        default="",
        help="Updated value for config entry. If not provided, the current config "
        "value is printed",
    ),
    interactive: bool = flag(["--interactive", "-i"]),
):
    """Read and write from the kslurm config"""
    config = appconfig.Config()
    if interactive:
        if entry not in INTERACTIVE_SETTINGS:
            raise InteractiveConfigError(f"{entry} can not be updated interactively")
        for key, new_value in INTERACTIVE_SETTINGS[entry]().items():
            if new_value is not None:
                config[key] = new_value
                continue
            if key in config:
                del config[key]
        config.write()
        return
    if not value:
        curr_value = config.get(entry)
        if curr_value is not None:
            print(curr_value)
        for key, value in config.get_children(entry):
            print(f"{key}: {value}")

    else:
        config[entry] = value
        config.write()



from __future__ import absolute_import, annotations

from rich.console import Console

from kslurm.style.themes import default

console = Console(theme=default)
stderr = Console(theme=default, stderr=True)



from __future__ import absolute_import, annotations

import uuid
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, TypeVar

from typing_extensions import ParamSpec

if TYPE_CHECKING:
    from typing_extensions import StrictTypeGuard
else:
    from typing_extensions import TypeGuard

    StrictTypeGuard = TypeGuard

import attr

import kslurm.args.actions as actions
import kslurm.args.matchers as matchers
from kslurm.args.arg import Arg, ParamSet, Parser
from kslurm.args.help_templates import PositionalArg, ShapeArg, SubcommandTemplate
from kslurm.args.protocols import Command, WrappedCommand
from kslurm.exceptions import ValidationError

T = TypeVar("T")
S = TypeVar("S")
P = ParamSpec("P")

Subcommand = tuple[str, WrappedCommand]


def positional(
    default: Optional[T] = None,
    *,
    help: str = "",
    name: str = "",
    format: Callable[[str], T] = actions.NO_CONVERT,
) -> Arg[T]:
    return Arg[T](
        parser=Parser(
            priority=0,
            match=matchers.everything(),
            action=actions.convert(format).replace(),
        ),
        help=help,
        help_template=PositionalArg(),
        name=name,
        default=default,
    )


def choice(
    match: List[str],
    *,
    default: Optional[T] = None,
    help: str = "",
    name: str = "",
    format: Callable[[str], T] = actions.NO_CONVERT,
) -> Arg[T]:
    def check_match(val: str) -> T:
        if val in match:
            return format(val)
        choices = "\n".join([f"\t• {m}" for m in match])
        raise ValidationError(f"Please select between:\n{choices}")

    choices = ", ".join(match)
    helptxt = help + "\n" if help else ""
    return Arg[T](
        parser=Parser(
            match=matchers.everything(),
            priority=0,
            action=actions.convert(check_match).replace(),
        ),
        help=f"{helptxt}options: [{choices}]",
        help_template=PositionalArg(),
        name=name,
        default=default,
    )


C = TypeVar("C", bound="Command[Any] | Command[[]]")


def _is_command(command: WrappedCommand | C) -> StrictTypeGuard[C]:
    return hasattr(command, "cli")


class InvalidSubcommand(Exception):
    pass


def subcommand(
    commands: Dict[str, WrappedCommand | Command[...] | Command[[]]],
    default: Optional[WrappedCommand | Command[...] | Command[[]]] = None,
):
    cli = {
        name: command.cli if _is_command(command) else command
        for name, command in commands.items()
    }

    _default = default.cli if default and _is_command(default) else default

    def get_action(val: str):
        if val in cli.keys():
            return (val, cli[val])
        choices = "\n".join([f"\t\t• {m}" for m in cli.keys()])
        err = f"Please select between:\n{choices}"
        if _default:
            raise InvalidSubcommand(err)
        raise ValidationError(err)

    return Arg(
        parser=Parser(
            match=matchers.everything(),
            priority=0,
            action=actions.convert(get_action).replace(),
            terminal=True,
        ),
        help_template=SubcommandTemplate(cli),
        name="subcommand",
        default=("", _default) if _default is not None else None,
    )


def shape(
    match: str,
    *,
    default: Optional[T] = None,
    format: Callable[[str], T] = actions.NO_CONVERT,
    help: str = "",
    name: str = "",
    syntax: str = "",
    examples: List[str] = [],
) -> Arg[T]:
    return Arg[T](
        parser=Parser(
            match=matchers.regex(match),
            priority=10,
            action=actions.convert(format).replace(),
        ),
        help=help,
        help_template=ShapeArg(
            syntax=syntax,
            examples=examples,
        ),
        name=name,
        default=default,
    )


def path(
    *,
    default: Optional[Path] = None,
    dir_only: bool = False,
    help: str = "",
    name: str = "",
) -> Arg[Path]:
    return Arg[Path](
        parser=Parser(
            match=matchers.path(is_dir=dir_only),
            priority=10,
            action=actions.convert(Path).replace(),
        ),
        help=help,
        name=name,
        default=default,
    )


def flag(
    match: list[str],
    default: Optional[bool] = False,
    help: str = "",
) -> Arg[bool]:

    return Arg[bool](
        parser=Parser(
            match=matchers.choice(*match).settings(duplicates=True),
            priority=20,
            action=actions.convert(bool).replace(),
        ),
        help=help,
        name=", ".join(match),
        default=default,
    )


def keyword(
    match: list[str],
    *,
    default: Optional[Any] = attr.NOTHING,
    format: Callable[[str], Any] = actions.NO_CONVERT,
    num: Optional[int] = None,
    help: str = "",
    lazy: bool = False,
    id: str = "",
) -> Any:
    pid = str(uuid.uuid4().int)
    action = (
        actions.convert(format).replace()
        if num is None
        else actions.convert(format).append()
    )
    return ParamSet(
        parent=Parser(
            id=pid,
            match=matchers.choice(*match).settings(duplicates=True),
            priority=20,
            action=actions.convert(bool).replace(),
        ),
        child=Parser(
            match=matchers.option_chain(pid, num or 1, matchers.everything()),
            priority=5 if lazy else 30,
            action=action,
        ),
        help=help,
        name=", ".join(match) + " <value>",
        default=default,
        id=id,
    )


class HelpRequest(Exception):
    pass


def help_parser() -> Parser[bool]:
    return Parser(
        match=matchers.choice("--help", "-h").settings(duplicates=True),
        priority=20,
        action=actions.raises(HelpRequest()),
        value=False,
    )



from __future__ import absolute_import, annotations

import json
import os
import re
import subprocess as sp
from collections import UserDict
from pathlib import Path
from typing import Any

from virtualenv.create import pyenv_cfg  # type: ignore

from kslurm.appconfig import PipDir
from kslurm.utils import get_hash


class KpyIndex(UserDict[str, str]):
    def __init__(self, slurm_tmpdir: Path):
        self._path = slurm_tmpdir / "tmp" / "kpy-index.json"
        if not self._path.exists():
            self._path.parent.mkdir(parents=True, exist_ok=True)
            with self._path.open("w") as f:
                json.dump({}, f)
                self.data = {}
        else:
            with self._path.open("r") as f:
                self.data = json.load(f)

    def write(self):
        with self._path.open("w") as f:
            json.dump(self.data, f)

    def __str__(self):
        return "• " + "\n• ".join(self.data.keys()) if self.data else ""


class VenvCache(UserDict[str, Path]):
    def __init__(self):
        pipdir = PipDir()
        self.venv_cache = pipdir / "venv_archives"
        self.venv_cache.mkdir(exist_ok=True)
        venvs_re = [
            re.search(r"(.+)\.tar\.gz$", str(f.name)) for f in self.venv_cache.iterdir()
        ]
        venvs = [r.group(1) for r in venvs_re if r]
        self.data = {v: self._construct_path(v) for v in venvs}

    def get_path(self, name: str):
        try:
            return self.data[name]
        except KeyError:
            return self._construct_path(name)

    def _construct_path(self, name: str):
        return (self.venv_cache / name).with_suffix(".tar.gz")

    def __str__(self):
        return "• " + "\n• ".join(self.data.keys()) if self.data else ""


def _pip_freeze(venv_dir: Path):
    return sp.run(
        [venv_dir / "bin" / "python", "-m", "pip", "freeze"], capture_output=True
    ).stdout


def _file_sub(path: Path, *replacements: tuple[str, str]):
    with path.open("r") as f:
        contents = f.read()
    sub = contents
    for replace in replacements:
        sub = re.sub(*replace, sub)
    with path.open("w") as f:
        f.write(sub)


class PromptRefreshError(Exception):
    pass


class VenvPrompt:
    def __init__(self, venv_dir: Path):
        self.cfg: Any = pyenv_cfg.PyEnvCfg.from_folder(venv_dir)  # type: ignore
        self.venv_dir = venv_dir
        try:
            self.name = self.cfg["prompt"]
        except KeyError:
            self.name = "venv"

    def update_prompt(self, name: str):
        self.name = name
        self.cfg["prompt"] = name

    def update_hash(self):
        self.cfg["state_hash"] = get_hash(_pip_freeze(self.venv_dir))

    def refresh(self):
        try:
            state_hash = self.cfg["state_hash"]
        except KeyError:
            raise PromptRefreshError()

        try:
            hsh = get_hash(_pip_freeze(self.venv_dir))
            if hsh != state_hash and self.name[0] != "*":
                self.update_prompt("*" + self.name)
            elif hsh == state_hash and self.name[0] == "*":
                self.update_prompt(self.name[1:])
            self.save()
        except KeyError:
            raise PromptRefreshError()

    def save(self):
        self.cfg.write()
        bin_dir = self.venv_dir / "bin"
        _file_sub(
            bin_dir / "activate",
            (
                r'\[\s"x[^"$]*"\s!=\sx\s\]',
                f'[ "x{self.name}" != x ]',
            ),
            (r'\sPS1="[^"$]*\$\{PS1\-\}"', f' PS1="({self.name}) ${{PS1-}}"'),
        )

        # _file_sub(
        #     r"\(\x27[^\x27]*\x27\s!=\s\x22\x22\)",
        #     f"('{self.name}' != \"\")",
        #     bin_dir / "activate.csh",
        # )

        # _file_sub(
        #     r"\(\x27[^\x27]*\x27\s!=\s\x22\x22\)",
        #     f"('{self.name}' != \"\")",
        #     bin_dir / "activate.csh",
        # )


def rebase_venv(venv_dir: Path):
    with (venv_dir / "bin" / "activate").open("r") as f:
        resolved = venv_dir.resolve()
        lines = f.read().splitlines()
        # \x22 and \x27 are " and ' char
        subbed = [
            re.sub(
                r"(?<=^VIRTUAL_ENV=([\x22\x27])).*(?=\1$)",
                str(resolved),
                line,
            )
            for line in lines
        ]
    with (venv_dir / "bin" / "activate").open("w") as f:
        f.write("\n".join(subbed))

    for root, _, files in os.walk(venv_dir / "bin"):
        for file in files:
            path = Path(root, file)
            if path.is_symlink() or not os.access(path, os.X_OK):
                continue
            with path.open("r") as f:
                lines = f.read().splitlines()
                try:
                    subbed = [
                        re.sub(
                            r"(?<=^#!)\/.*python.*$",
                            str(resolved / "bin" / "python"),
                            lines[0],
                        ),
                        *lines[1:],
                    ]
                except IndexError:
                    subbed = lines
            with path.open("w") as f:
                f.write("\n".join(subbed))



# from __future__ import absolute_import

# from typing import Any, List, NamedTuple

from __future__ import absolute_import, annotations

# import pytest
# from _pytest.fixtures import SubRequest
# import kslurm.args.helpers as helpers
# import kslurm.args.parser as sc
# from kslurm.args import Arg
# from kslurm.exceptions import CommandLineErrorGroup

# from .models import models
# from .models.common import ModelTest


# class ModelCase(NamedTuple):
#     model: object
#     test: List[str]
#     outcome: List[Arg[Any]]


# @pytest.fixture(scope="module", params=models)
# def model(request: SubRequest):  # type: ignore
#     return request.param  # type: ignore


# def test_relabelled_args_convert_back_to_model(model: ModelTest):
#     labelled = helpers.get_arg_list(model.model)
#     remodelled = helpers.update_model(labelled, model.model)
#     assert remodelled == model.model


# def test_parse_args(model: ModelTest):
#     for args, outcome in zip(model.tests, model.expected_outcomes):
#         if isinstance(outcome[0], Exception):
#             try:
#                 sc.parse_args(args, model.model, True)
#             except CommandLineErrorGroup as err_group:
#                 assert [err.msg for err in outcome] == [  # type: ignore
#                     err.msg for err in err_group.sub_errors
#                 ]
#             else:
#                 assert False
#         else:
#             parsed_model = sc.parse_args(args, model.model, True)
#             labelled_list = helpers.get_arg_list(parsed_model)

#             assert set(labelled_list) == set(outcome)  # type: ignore



from __future__ import absolute_import, annotations

import itertools as it
import re
import typing
from collections import defaultdict
from typing import (
    Any,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    TypeVar,
    Union,
    cast,
    overload,
)

import attr
import docstring_parser as docparse
import more_itertools as itx

from kslurm.args import actions
from kslurm.args.arg import (
    Arg,
    Helpable,
    ParamInterface,
    ParamSet,
    Parser,
    SimpleParsable,
)
from kslurm.exceptions import CommandLineError

T = TypeVar("T")
S = TypeVar("S")
C = TypeVar("C", bound=Arg[Any])

ModelType = type[Any]
ModelDict = dict[str, ParamInterface[Any]]


def group_by_type(items: Iterable[T]) -> Dict[type, List[T]]:
    groupedDict = cast(DefaultDict[type, List[T]], defaultdict(list))
    for key, value in it.groupby(items, type):
        groupedDict[key] += list(value)
    return groupedDict


def get_parsers(models: ModelDict):
    result: dict[str, Parser[Any]] = {}
    for param in models.values():
        result.update(param.get_parsers())
    return result


@overload
def read_parsers(
    models: ModelDict,
    parsers: dict[str, Parser[Any]],
    raw_values: Literal[False] = ...,
    updated: Literal[False] = ...,
) -> tuple[dict[str, Any], dict[str, CommandLineError]]:
    ...


@overload
def read_parsers(
    models: ModelDict,
    parsers: dict[str, Parser[Any]],
    raw_values: Literal[True] = ...,
    updated: Literal[False] = ...,
) -> tuple[dict[str, Any], dict[str, CommandLineError], dict[str, list[str]]]:
    ...


@overload
def read_parsers(
    models: ModelDict,
    parsers: dict[str, Parser[Any]],
    raw_values: Literal[False] = ...,
    updated: Literal[True] = ...,
) -> tuple[dict[str, Any], dict[str, CommandLineError], dict[str, bool]]:
    ...


@overload
def read_parsers(
    models: ModelDict,
    parsers: dict[str, Parser[Any]],
    raw_values: Literal[True] = ...,
    updated: Literal[True] = ...,
) -> tuple[
    dict[str, Any], dict[str, CommandLineError], dict[str, list[str]], dict[str, bool]
]:
    ...


@overload
def read_parsers(
    models: ModelDict,
    parsers: dict[str, Parser[Any]],
    raw_values: bool = ...,
    updated: bool = ...,
) -> tuple[dict[str, Any], ...]:
    ...


def read_parsers(
    models: ModelDict,
    parsers: dict[str, Parser[Any]],
    raw_values: bool = False,
    updated: bool = False,
) -> Any:
    default: list[dict[str, Any]] = [{}, {}]
    if raw_values:
        default.append({})
    if updated:
        default.append({})
    unmerged = [
        model.read_parsers(parsers, raw_values, updated) for model in models.values()
    ]
    result = [_merge_dicts(elem) for elem in zip(*unmerged)]
    return tuple(result) or tuple(default)


@attr.frozen
class ParamAnnotation:
    is_optional: bool
    is_list: bool
    is_generic: bool
    main_type: type[Any]
    root_type: type[Any]

    @classmethod
    def parse(cls, __t: Optional[type[Any]], /):
        optional = False
        is_generic = False
        is_list = False
        if typing.get_origin(__t) == Union and type(None) in (
            args := typing.get_args(__t)
        ):
            i = args.index(type(None))
            rest = (*args[:i], *args[i + 1 :])
            if len(rest) > 1 or not len(rest):
                raise TypeError()
            optional = True
            __t = rest[0]
        main_type = __t
        if len(typing.get_args(__t)):
            is_generic = True
        if typing.get_origin(__t) == list:
            is_list = True
            __t = typing.get_args(__t)[0]
        if __t is None:
            __t = Literal[None]
        if main_type is None:
            main_type = Literal[None]
        return cls(
            is_optional=optional,
            is_generic=is_generic,
            is_list=is_list,
            main_type=main_type,
            root_type=__t,
        )


def _at_least(__iter: Iterable[T], __n: int, __default: T, /):
    i = 0
    for el in __iter:
        i += 1
        yield el
    yield from it.repeat(__default, max(__n - 1, 0))


def _parse_description(desc: str):
    x = [
        (match.group(1), match.group(2))
        if (match := re.match(r"@([\w\.]+)[ ]+(.*)?$", line))
        else (None, line)
        for line in filter(None, desc.splitlines())
    ]
    main, *_params = _at_least(itx.split_before(x, lambda _: _[0], maxsplit=1), 1, None)
    params = itx.first(_params, None)
    if params and itx.ilen(filter(lambda _: _[0] is None, params)):
        raise Exception()
    main = cast("list[tuple[None, str]] | None", main)
    params = cast("list[tuple[str, str]] | None", params)
    helptxt = " ".join([s[1] for s in main or []])
    meta: dict[str, Any] = {}
    for keypath, value in params or []:
        pointer = meta
        *keys, final = keypath.split(".")
        for key in keys:
            if key not in pointer:
                pointer[key] = {}
            pointer = pointer[key]
        pointer[final] = value
    return helptxt, meta


def get_arg_dict(models: ModelType) -> ModelDict:
    if not attr.has(models):
        raise TypeError(f"{type(models)} is not a supported object for arg models")

    doc = docparse.parse(models.__doc__ or "")
    arghelps = {arg.arg_name: arg for arg in doc.params}
    fields = attr.fields(models)
    result: ModelDict = {}
    for field in fields:
        if field.type is attr.NOTHING:
            raise TypeError(f"{field.name} must be annotated")

        annotation = ParamAnnotation.parse(field.type)

        # temporary assertion until we handle other defaults
        assert isinstance(field.default, SimpleParsable)

        default: SimpleParsable[Any] = field.default

        action = default.primary_parser.action
        if not action.has_formatter:
            newtype = annotation.root_type

            if annotation.is_list and not (
                isinstance(default, ParamSet) and isinstance(action, actions.append)
            ):
                raise TypeError(
                    f"invalid annotation for '{field.name}' in class "
                    f"'{models.__name__}': list annotations are not currently "
                    "supported in this context. Please provide a formatter to convert "
                    f"your arguments into lists (current annotation: '{field.type}')"
                )

            if not callable(newtype):
                raise TypeError(
                    f"{field.name} in '{models.__qualname__}' must be annotated with a "
                    "callable type"
                )

            default = default.with_primary_parser(
                attr.evolve(
                    default.primary_parser, action=action.with_converter(newtype)
                )
            )

        updates: dict[str, Any] = {"id": field.name}
        if isinstance(default, ParamSet) and default.default is attr.NOTHING:
            updates["default"] = annotation.root_type()

        if annotation.is_optional:
            updates["optional"] = True

        if field.name in arghelps:
            desc, meta = _parse_description(arghelps[field.name].description or "")
            if "name" in meta:
                updates["name"] = meta["name"]
            if isinstance(default, Helpable):
                updates["help"] = desc
                if default.help_template and "help" in meta:
                    try:
                        updates["help_template"] = default.help_template.update_meta(
                            **meta["help"]
                        )
                    except TypeError as err:
                        raise CommandLineError(
                            f"Unexpected param found when parsing {models} docstring, "
                            f"attribute '{field.name}': " + err.args[0]
                        )

        assert isinstance(default, ParamInterface)
        result[field.name] = attr.evolve(default, **updates)
    return result


def finalize_model(
    arg_updates: ModelDict, model: ModelType, exclude: list[str] = []
) -> Any:
    update = {
        arg.id: arg.value for arg in arg_updates.values() if arg.id not in exclude
    }
    if isinstance(model, dict):
        return cast(T, {**model, **update})
    elif attr.has(model):
        return model(**update)
    else:
        raise Exception(f"{type(model)} is not a supported object for arg models")


def get_help(model: Any) -> bool:
    if attr.has(model):
        help = getattr(model, "help", None)
        return getattr(help, "value", False)
    if isinstance(model, dict):
        help: Any = model.get("help", None)  # type: ignore
        return getattr(help, "value", False)
    raise Exception(f"{type(model)} is not a supported type for arg models")


def _merge_dicts(dicts: Iterable[dict[T, S]]) -> dict[T, S]:
    result: dict[T, S] = {}
    for d in dicts:
        result.update(d)
    return result



from __future__ import absolute_import

from pathlib import Path

import attr

import kslurm.models.formatters as formatters
import kslurm.models.validators as validators
from kslurm.args import flag, keyword, path, shape


@attr.s(auto_attribs=True)
class SlurmModel:
    """
    Attributes:
        time (int):
            Amount of time requested. Written as [days-]hr:min
            @help.syntax [d-]dd:dd
            @name Time

        gpu:
            Request a gpu instance

        cpu:
            Number of CPUs to request. Maximum of 32 on Graham Nodes
            @help.syntax d
            @name Number of CPUs

        mem:
            Amount of memory to request
            @help.syntax d(M|G)[B]
            @name Memory

        x11:
            Request X-forwarding for GUI applications. Must also request X-forwarding on
            your SSH connection.

        directory:
            Change working directory before submitting the command. All other relative
            paths will be evaluated relative to the new directory.
            @name Directory

        test:
            Test mode. Will echo SLURM settings and the command without submitting
            anything.

        job_template:
            Set CPU, memory, and time from a template. Run with -J to see list of
            templates. Any template item can be overridden by supplying the appropriate
            arg.

        list_job_templates:
            List all available templates.

        account:
            Compute account to submit the job under.

        venv:
            kpy venv to load
    """

    time: int = shape(
        match=r"^([0-9]{1,2}-)?[0-9]{1,2}:[0-9]{2}$",
        format=formatters.time,
        default=180,
        examples=["3:00 (3hr)", "1-00:00 (1 day)"],
    )

    gpu: bool = flag(match=["gpu"])

    cpu: int = shape(
        match=r"^[0-9]+$",
        default=1,
        examples=["1 (1 cpu)", "16 (16 cpus)"],
    )

    mem: int = shape(
        match=r"^[0-9]+[KMGkmg][Ii]?[Bb]?$",
        format=formatters.mem,
        default=4000,
        examples=["3000MB (3 GB)", "16G (16GB)"],
    )

    x11: bool = flag(match=["x11", "--x11"])

    directory: Path = path(
        default=Path(),
        dir_only=True,
    )

    test: bool = flag(match=["-t", "--test"])

    job_template: str = keyword(
        match=["-j", "--job-template"], format=validators.job_template
    )

    list_job_templates: bool = flag(match=["-J", "--list", "-l"])

    account: str = keyword(match=["-a", "--account"])

    venv: str = keyword(["--venv"], default="")



from __future__ import absolute_import

import importlib.resources as impr
import json
import os
import re
import signal
import subprocess as sp
import sys
import tempfile as tmp
from pathlib import Path
from typing import Any, Union

import attrs
from yaspin import yaspin

import kslurm.bin
import kslurm.text as txt
from kslurm.args import HelpRequest, InvalidSubcommand, Subcommand, subcommand
from kslurm.args.arg_types import choice, flag, keyword
from kslurm.args.command import CommandError, Parsers, command
from kslurm.args.help import HelpText
from kslurm.exceptions import TemplateError
from kslurm.models.slurm import SlurmModel
from kslurm.slurm.slurm_command import SlurmCommand
from kslurm.style import console
from kslurm.venv import VenvCache


@attrs.frozen
class _KjupyterEnv:
    active: bool
    logs: Path
    domain: str = ""
    hostname: str = ""
    username: str = ""
    port: str = ""
    token: str = ""
    url: str = ""

    @staticmethod
    def get_env_var(name: str):
        return "_KJUPYTER_" + name

    def export(self):
        for field, value in attrs.asdict(self).items():
            os.environ[self.get_env_var(field)] = str(value or "")

    def set_server_vals(
        self, hostname: str, username: str, port: str, token: str, url: str, domain: str
    ):
        return attrs.evolve(
            self,
            hostname=hostname,
            username=username,
            port=port,
            token=token,
            url=url,
            domain=domain,
        )

    @classmethod
    def load(cls):
        fields = attrs.fields(cls)
        vals = {
            field.name: field.type(val)
            if (val := os.environ.get(cls.get_env_var(field.name))) is not None
            else field.type()
            for field in fields
            if field.type is not attrs.NOTHING and field.type is not None
        }
        return cls(**vals)


@attrs.frozen
class _KjupyterExtendedModel(SlurmModel):
    debug: bool = flag(["--debug"])


_KjupyterExtendedModel.__doc__ = SlurmModel.__doc__


def _get_tunnel_script(*, port: str, domain: str, username: str, hostname: str):
    return f"ssh -L {port}:{hostname}:{port} {username}@{domain}"


def _get_browser_url(*, port: str, token: str):
    return f"http://localhost:{port}/lab?token={token}"


@command(terminate_on_unknown=True, allow_unknown=False)
def _kjupyter(
    args: Union[_KjupyterExtendedModel, TemplateError],
    arglist: Parsers,
):
    """Start a Jupyter session

    Begins a Jupyter session on an interactive node. By default, it requests 3hr. No
    more time than this should be requested, as this could significantly delay the
    start of the server.

    The --venv arg can be used to specify saved venv. This venv must already
    have jupyter-lab installed, or the command will fail. This will install the
    venv into local scratch, and will thus lead to the best performance.
    Alternatively, you can run the in an already activated virtualenv containing
    jupyter-lab.
    """
    slurm = SlurmCommand(args, [], arglist)
    slurm.name = "jupyter"
    assert isinstance(args, SlurmModel)

    if args.venv:
        venv_cache = VenvCache()
        if args.venv not in venv_cache:
            print("Valid venvs:\n" + str(venv_cache))
            return 1

    env = _KjupyterEnv(active=True, logs=Path(tmp.mkstemp(prefix="kjupyter_logs.")[1]))
    env.export()

    cmd = [
        "jupyter-lab",
        "--ip",
        "$(hostname -f)",
        "--no-browser",
        "-y",
        "2>&1",
        "|",
        "tee",
        str(env.logs),
    ]
    console.print(
        txt.KRUN_CMD_MESSAGE.format(args=slurm.slurm_args, command=" ".join(cmd))
    )
    # venv_load = ["source", str(path), "; kpy load", args.venv] if args.venv else []
    with impr.path(kslurm.bin, "kpy-wrapper.sh") as token:
        slurm.command = (
            [
                "command -v jupyter-lab > /dev/null || (echo 'jupyter-lab not found' "
                "&& pip install jupyterlab);",
                *cmd,
            ]
            if args.venv
            else cmd
        )

    if slurm.test:
        print(slurm.run)
        return

    proc = sp.Popen(
        slurm.run,
        shell=True,
        stdin=sp.PIPE,
        stdout=sp.PIPE,
        stderr=sp.STDOUT,
    )
    jobid = "0"

    def signal_handler(*args: Any):
        sp.run(f"scancel {jobid}", shell=True)
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)

    port = None
    with yaspin(text="Acquiring resources") as spinner:
        if args.debug:
            spinner.stop()
        while proc.poll() is None:
            if proc.stdout:
                line = proc.stdout.readline().decode().strip()
                queued = re.match(
                    r"^srun: job (\d+) queued and waiting for resources", line
                )
                url = re.match(
                    r"^https?:\/\/((?:[\w-]+\.)+(?:[\w]+)):(\d+)\/lab\?token=(\w+)$",
                    line,
                )

                # unused code to grab the address lines
                _ = any(
                    [
                        re.match(
                            r"^To access the server, open this file in a browser:$",
                            line,
                        ),
                        re.match(r"^file:\/\/\/(?:[\w.\-_]+\/)+[\w.\-_]+\.html$", line),
                        re.match(r"^Or copy and paste one of these URLs:$", line),
                        re.match(
                            r"^or http:\/\/(?:[\d]+\.)+\d+:\d+\/lab\?token=\w+$",
                            line,
                        ),
                    ]
                )

                if queued:
                    jobid = queued.group(1)

                elif re.match(r"^Unpacking venv \'.+\'", line):
                    spinner.text = "Unpacking venv"
                    if args.debug:
                        print(line)

                elif re.match(r"jupyter-lab not found", line):
                    spinner.text = "Installing jupyter-lab"
                    if args.debug:
                        print(line)

                elif url:
                    hostname = url.group(1)
                    port = url.group(2)
                    token = url.group(3)
                    username = sp.getoutput("whoami").strip()
                    try:
                        domain = (
                            sp.check_output(["wget", "-qO-", "ipinfo.io/ip"])
                            .decode()
                            .strip()
                        )
                    except (RuntimeError, sp.CalledProcessError):
                        domain = "<hostname>"
                    env.set_server_vals(
                        hostname=hostname,
                        username=username,
                        port=port,
                        token=token,
                        url=url.group(0),
                        domain=domain,
                    ).export()
                    spinner.text = "Finished!"
                    if not args.debug:
                        spinner.ok("🚀")
                        console.print(
                            txt.JUPYTER_WELCOME.format(
                                browser_url=_get_browser_url(port=port, token=token),
                                tunnel_script=_get_tunnel_script(
                                    port=port,
                                    domain=domain,
                                    username=username,
                                    hostname=hostname,
                                ),
                                url=url.group(0),
                            )
                        )
                        break

                elif args.debug:
                    print(line)

        assert port
        sp.run(
            [
                "srun",
                f"--jobid={jobid}",
                "--pty",
                "--overlap",
                "bash",
                "-c",
                slurm.venv_activate + " bash",
            ]
        )
        with yaspin(text="Shutting down jupyter") as spinner:
            sp.run(
                [
                    "srun",
                    f"--jobid={jobid}",
                    "--overlap",
                    "bash",
                    "--noprofile",
                    "--norc",
                    "-c",
                    slurm.venv_activate + f" jupyter lab stop {port}",
                ],
                capture_output=True,
            )
            spinner.ok()


@command(inline=True)
def _log(lines: int = keyword(["-n", "--lines"], default=20)):
    """Display the logs from the jupyter session

    Args:
        lines: Number of lines to display (default 20)
    """
    env = _KjupyterEnv.load()
    lines_arg = ["--lines", str(lines)]
    proc = sp.Popen(["tail", env.logs, *lines_arg], stdout=sp.PIPE)
    while proc.poll() is None:
        if proc.stdout:
            for line in proc.stdout.readlines():
                console.print(line.decode().rstrip(), markup=False)


@command(inline=True)
def _console():
    """Begin interactive python console

    The jupyter-console package must already be installed in the virtual env. A kernel
    must already be started by a notebook file
    """
    env = _KjupyterEnv.load()
    sessions = json.loads(
        sp.check_output(
            [
                "curl",
                "-sSLG",
                f"{env.hostname}:{env.port}/api/sessions",
                "--data-urlencode",
                f"token={env.token}",
            ]
        )
    )
    if not len(sessions):
        raise CommandError(
            "No sessions found. Activate a jupyter notebook before running this "
            "command."
        )
    sp.run(["jupyter", "console", "--existing"])


@command(inline=True)
def _url(
    format: str = choice(
        ["browser", "server"],
    )
):
    """Echo the jupyter server url

    Args:
        format: The url format to retrieve: browser for the webapp, and server for
            vscode
    """
    env = _KjupyterEnv.load()
    if format == "browser":
        print(_get_browser_url(port=env.port, token=env.token))
        return 0
    print(env.url)


@command(inline=True)
def _tunnel():
    """Echo the bash code to open an ssh tunnel to the jupyter server"""
    env = _KjupyterEnv.load()
    print(
        _get_tunnel_script(
            port=env.port,
            domain=env.domain,
            username=env.username,
            hostname=env.hostname,
        )
    )


@attrs.frozen
class _JupyterModel:
    cmd: Subcommand = subcommand(
        {"log": _log, "console": _console, "url": _url, "tunnel": _tunnel},
        default=_kjupyter,
    )


@command
def kjupyter(
    entrypoint: str,
    args: Union[_JupyterModel, InvalidSubcommand, HelpRequest],
    command_args: list[str],
    helptext: HelpText,
):
    """Helper commands to manage jupyter session

    To close the jupyter server, press CTRL-D or run `exit`
    """
    env = _KjupyterEnv.load()
    if isinstance(args, HelpRequest):
        if env.active:
            console.print(helptext)
        else:
            console.print(_kjupyter.get_helptext(entrypoint))
        return
    if isinstance(args, InvalidSubcommand):
        if env.active:
            console.print(args.args[0])
            return
        return _kjupyter.cli(["kjupyter", *command_args])
    name, func = args.cmd
    if not name:
        return func([name, *command_args])
    entry = f"{entrypoint} {name}"
    if not env.active:
        raise CommandError(
            f"The '{name}' command is only available in a running jupyter notebook. "
            "Use the kjupyter base command to start a new notebook session"
        )
    return func([entry, *command_args])


if __name__ == "__main__":
    # KjupyterEnv(True, Path("./poetry.lock")).export()
    kjupyter.cli(["kjupyter", "url", "--help"])



from __future__ import absolute_import, annotations

import abc
import textwrap as txt
from typing import (
    Any,
    Callable,
    DefaultDict,
    Generic,
    Literal,
    Optional,
    Protocol,
    TypeVar,
    Union,
    overload,
)

import attr
from colorama import Fore, Style
from rich.table import Table
from rich.text import Text
from typing_extensions import Self

from kslurm.exceptions import CommandLineError, MandatoryArgError, ValidationError

T = TypeVar("T")
S = TypeVar("S")

HelpRow = Union[list[Union[Text, str]], list[Text], list[str]]


class _SkipHelp:
    """
    Sentinel class to indicate argument should not be shown in help messages

    ``_SkipHelp`` is a singleton. There is only ever one of it.

    """

    _singleton = None

    def __new__(cls):
        if _SkipHelp._singleton is None:
            _SkipHelp._singleton = super(_SkipHelp, cls).__new__(cls)
        return _SkipHelp._singleton

    def __repr__(self):
        return "SKIPHELP"

    def __bool__(self):
        return False

    def __len__(self):
        return 0  # __bool__ for Python 2

    pass


SKIPHELP: Any = _SkipHelp()


def _mandatoryarg_err(
    label: str,
    help: str | None = None,
    help_template: AbstractHelpTemplate | None = None,
):
    _h = help or help_template.help if help_template else None
    return MandatoryArgError(
        f"{Fore.RED + Style.BRIGHT}ERROR:{Style.RESET_ALL}\n"
        + txt.indent(
            f"{Style.BRIGHT}'{label}'{Style.RESET_ALL} has not been "
            "provided a value.\n",
            "    ",
        )
        + (
            f"{Fore.CYAN + Style.BRIGHT}HELP:{Style.RESET_ALL}\n"
            + txt.indent(f"{_h}", "    ")
            if _h
            else ""
        )
    )


class AbstractHelpTemplate(abc.ABC):
    title: str
    description: str | None = None
    header: list[str]
    cls_usage: str = ""
    rows: dict[str, list[HelpRow]] = DefaultDict(list)
    priority: int = 0
    right_align_cols: int = 1

    def usage(self, name: str, help: str, default: Optional[str]) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def update_meta(self, **kwargs: Any) -> Self:
        ...

    @abc.abstractmethod
    def row(
        self, name: str, help: str, default: Optional[str]
    ) -> Union[list[HelpRow], HelpRow, None]:
        ...

    @abc.abstractproperty
    def help(self) -> str:
        ...

    def add_row(self, name: str, help: str, default: Optional[str]) -> None:
        new_row = self.row(name, help, default)
        if new_row:
            if all([isinstance(row, list) for row in new_row]):
                self.rows[self.__class__.__name__].extend(new_row)  # type: ignore
            elif all([not isinstance(row, list) for row in new_row]):
                self.rows[self.__class__.__name__].append(new_row)  # type: ignore
            else:
                raise TypeError(
                    f"row() in {self} must not return mixture of lists and items"
                )

    @classmethod
    def table(cls):
        t = Table.grid(padding=(0, 2), expand=True)
        for _ in range(cls.right_align_cols):
            t.add_column(justify="right", no_wrap=True)
        if cls.header:
            t.add_row(*cls.header, style="bold")
        for r in cls.rows[cls.__name__]:
            t.add_row(*list(r))
        return t


@attr.frozen
class Context:
    args: list[str]
    current_arg: int
    params: dict[str, Parser[Any]]
    last_matched: Optional[Parser[Any]]


class ActionProtocol(Generic[T], Protocol):
    @property
    def has_formatter(self) -> bool:
        ...

    def with_converter(self, _converter: Callable[[str], T], /) -> Self:
        ...

    def __call__(self, arg: str, parser: Parser[Any], context: Context) -> T:
        ...


@attr.frozen
class Parser(Generic[T]):
    priority: int
    match: Callable[[str, Self, Context], Union[bool, str]]
    action: ActionProtocol[T]
    id: str = ""
    terminal: bool = False
    value: Optional[T] = None
    raw_value: list[str] = []
    validation_err: Optional[ValidationError] = None

    def with_id(self, id: str):
        return attr.evolve(self, id=id)

    def with_value(self, value: Optional[str], context: Context):
        if value is None:
            return attr.evolve(self, value=value, values=[])

        try:
            return attr.evolve(
                self,
                value=self.action(value, self, context),
                raw_value=self.raw_value + [value],
            )
        except ValidationError as err:
            return attr.evolve(
                self,
                validation_err=ValidationError(
                    f"""{Fore.RED + Style.BRIGHT}ERROR:{Style.RESET_ALL}
    Invalid value for "{Style.BRIGHT}{{label}}{Style.RESET_ALL}": {err.msg}"""
                ),
            )

    def __repr__(self) -> str:
        # values_repr =
        # " : '" + "', '".join(map(str, self.values)) if self.values else ""
        return f"<{self.__class__.__name__}: '{self.id}' = {self.value}'>"


@attr.frozen
class ParamInterface(abc.ABC, Generic[T]):
    id: str
    name: str

    @property
    @abc.abstractmethod
    def value(self) -> T:
        ...

    @abc.abstractmethod
    def get_parsers(self) -> dict[str, Parser[Any]]:
        ...

    @abc.abstractmethod
    @overload
    def read_parsers(
        self,
        parsers: dict[str, Parser[Any]],
        raw_values: Literal[False] = ...,
        updated: Literal[False] = ...,
    ) -> tuple[dict[str, Any], dict[str, CommandLineError]]:
        ...

    @abc.abstractmethod
    @overload
    def read_parsers(
        self,
        parsers: dict[str, Parser[Any]],
        raw_values: Literal[True] = ...,
        updated: Literal[False] = ...,
    ) -> tuple[dict[str, Any], dict[str, CommandLineError], dict[str, list[str]]]:
        ...

    @abc.abstractmethod
    @overload
    def read_parsers(
        self,
        parsers: dict[str, Parser[Any]],
        raw_values: Literal[False] = ...,
        updated: Literal[True] = ...,
    ) -> tuple[dict[str, Any], dict[str, CommandLineError], dict[str, bool]]:
        ...

    @abc.abstractmethod
    @overload
    def read_parsers(
        self,
        parsers: dict[str, Parser[Any]],
        raw_values: Literal[True] = ...,
        updated: Literal[True] = ...,
    ) -> tuple[
        dict[str, Any],
        dict[str, CommandLineError],
        dict[str, list[str]],
        dict[str, bool],
    ]:
        ...

    @abc.abstractmethod
    @overload
    def read_parsers(
        self,
        parsers: dict[str, Parser[Any]],
        raw_values: bool = ...,
        updated: bool = ...,
    ) -> tuple[dict[str, Any], ...]:
        ...

    @abc.abstractmethod
    def read_parsers(
        self,
        parsers: Any,
        raw_values: Any = False,
        updated: Any = False,
    ) -> Any:
        ...

    def _get_parser_extras(
        self,
        parsers: dict[str, Parser[Any]],
        raw_values: bool,
        updated: bool,
    ) -> Any:
        raw_vals = (
            [{id: parser.raw_value for id, parser in parsers.items()}]
            if raw_values
            else []
        )
        updated_vals = (
            [{id: parser.value is not None for id, parser in parsers.items()}]
            if updated
            else []
        )
        return (*raw_vals, *updated_vals)

    @property
    def label(self) -> str:
        if not self.name:
            return self.id
        return self.name


class SimpleParsable(abc.ABC, Generic[T]):
    @property
    @abc.abstractmethod
    def primary_parser(self) -> Parser[T]:
        ...

    @abc.abstractmethod
    def with_primary_parser(self, parser: Parser[T]) -> Self:
        ...


class Helpable(abc.ABC):
    help: str
    help_template: Optional[AbstractHelpTemplate]


@attr.frozen(eq=False)
class Arg(ParamInterface[T], SimpleParsable[T], Helpable):
    parser: Parser[T]
    default: Optional[T] = None
    id: str = ""
    raw_value: list[str] = []
    help: str = ""
    help_template: Optional[AbstractHelpTemplate] = None
    name: str = ""
    optional: bool = False

    _value: Optional[T] = None

    @property
    def value(self) -> T:
        if self._value is not None:
            return self._value

        if self.default is not None:
            return self.default

        raise _mandatoryarg_err(
            self.label,
            self.help,
            self.help_template,
        )

    @property
    def assigned_value(self) -> Optional[T]:
        return self._value

    def get_parsers(self):
        return {self.id: self.parser.with_id(self.id)}

    def read_parsers(
        self,
        parsers: dict[str, Parser[Any]],
        raw_values: bool = False,
        updated: bool = False,
    ):
        parser = parsers[self.id]
        err = parser.validation_err
        if err is not None:
            err.format(label=self.label)
        if self.optional or parser.value is not None:
            value = {self.id: parser.value}

        elif self.optional:
            value = {self.id: None}

        elif self.default is not None:
            value = {self.id: self.default}
        else:
            value = {}
            err = err or _mandatoryarg_err(self.label, self.help, self.help_template)

        err_dict = {parser.id: err} if err is not None else {}
        extras = self._get_parser_extras(parsers, raw_values, updated)
        return value, err_dict, *extras

    @property
    def primary_parser(self) -> Parser[T]:
        return self.parser

    def with_primary_parser(self, parser: Parser[T]) -> Self:
        return attr.evolve(self, parser=parser)

    def __repr__(self) -> str:
        # values_repr =
        # " : '" + "', '".join(map(str, self.values)) if self.values else ""
        return f"<{self.__class__.__name__}: '{self.id}' = {self._value}'>"

    def __eq__(self, o: object) -> bool:
        if isinstance(o, self.__class__):
            return o._value == self._value and o.id == self.id
        else:
            return False

    def with_id(self, id: str):
        return attr.evolve(self, id=id)


@attr.frozen
class ParamSet(ParamInterface[T], SimpleParsable[T], Helpable):
    parent: Parser[bool]
    child: Parser[T]
    default: Optional[T] = None
    id: str = ""
    name: str = ""
    help: str = ""
    help_template: Optional[AbstractHelpTemplate] = None
    optional: bool = False

    @property
    def value(self):
        if self.child.value is not None:
            return self.child.value

        if self.default is not None:
            return self.default

        raise _mandatoryarg_err(self.label, self.help, self.help_template)

    @property
    def assigned_value(self):
        return self.child.value

    @property
    def raw_value(self):
        return self.child.raw_value

    def get_parsers(self):
        return {self.parent.id: self.parent, self.id: self.child.with_id(self.id)}

    def read_parsers(
        self,
        parsers: dict[str, Parser[Any]],
        raw_values: bool = False,
        updated: bool = False,
    ):
        child = parsers[self.id]
        err = child.validation_err
        if child.value is not None:
            value = {child.id: child.value}

        elif self.default is not None:
            value = {child.id: self.default}

        elif self.optional:
            value = {child.id: None}

        else:
            value = {}
            err = err or _mandatoryarg_err(self.label, self.help, self.help_template)

        err_dict = {child.id: err} if err is not None else {}
        extras = self._get_parser_extras(parsers, raw_values, updated)

        return value, err_dict, *extras

    @property
    def primary_parser(self) -> Parser[T]:
        return self.child

    def with_primary_parser(self, parser: Parser[T]) -> Self:
        return attr.evolve(self, child=parser)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: '{self.id}' = {self.assigned_value}'>"



from __future__ import absolute_import

from pathlib import Path
from typing import Any, Callable, Literal, NoReturn, Optional, TypeVar, overload

from typing_extensions import ParamSpec

import kslurm.args.arg_types as arg_types
from kslurm.args.arg import ParamSet, Parser
from kslurm.args.command import Command
from kslurm.args.protocols import WrappedCommand
from kslurm.exceptions import ValidationError

T = TypeVar("T")
P = ParamSpec("P")

Subcommand = tuple[str, WrappedCommand]

@overload
def positional(
    default: Literal[None] = ...,
    *,
    help: str = ...,
    name: str = ...,
) -> Any: ...
@overload
def positional(
    default: Literal[None] = ...,
    *,
    format: Callable[[str], T],
    help: str = ...,
    name: str = ...,
) -> T: ...
@overload
def positional(
    default: T,
    *,
    format: Callable[[str], T] = ...,
    help: str = ...,
    name: str = ...,
) -> T: ...
@overload
def choice(
    match: list[str],
    *,
    default: Literal[None] = ...,
    help: str = ...,
    name: str = ...,
) -> Any: ...
@overload
def choice(
    match: list[str],
    *,
    default: Literal[None] = ...,
    format: Callable[[str], T],
    help: str = ...,
    name: str = ...,
) -> T: ...
@overload
def choice(
    match: list[str],
    *,
    default: T,
    format: Callable[[str], T] = ...,
    help: str = ...,
    name: str = ...,
) -> T: ...
def subcommand(
    commands: dict[str, WrappedCommand | Command[...] | Command[[]]],
    default: Optional[WrappedCommand | Command[...] | Command[[]]] = ...,
) -> arg_types.Subcommand: ...

class InvalidSubcommand(ValidationError):
    pass

@overload
def shape(
    match: str,
    *,
    default: None = ...,
    help: str = ...,
    name: str = ...,
    syntax: str = ...,
    examples: list[str] = ...,
) -> Any: ...
@overload
def shape(
    match: str,
    *,
    default: None = ...,
    help: str = ...,
    format: Callable[[str], T],
    name: str = ...,
    syntax: str = ...,
    examples: list[str] = ...,
) -> T: ...
@overload
def shape(
    match: str,
    *,
    default: T,
    help: str = ...,
    format: Callable[[str], T] = ...,
    name: str = ...,
    syntax: str = ...,
    examples: list[str] = ...,
) -> T: ...
def path(
    *,
    default: Optional[Path] = ...,
    dir_only: bool = ...,
    help: str = ...,
    name: str = ...,
) -> Path: ...
def flag(
    match: list[str],
    default: Optional[bool] = ...,
    help: str = ...,
) -> bool: ...
@overload
def keyword(
    match: list[str],
    *,
    default: None = ...,
    num: None = ...,
    lazy: bool = ...,
    help: str = ...,
) -> Any: ...
@overload
def keyword(
    match: list[str],
    *,
    default: None = ...,
    num: int,
    lazy: bool = ...,
    help: str = ...,
) -> list[Any]: ...
@overload
def keyword(
    match: list[str],
    *,
    default: T,
    format: Callable[[str], T] = ...,
    num: None = ...,
    lazy: bool = ...,
    help: str = ...,
) -> T: ...
@overload
def keyword(
    match: list[str],
    *,
    default: Optional[Any] = ...,
    format: Callable[[str], Any] = ...,
    num: Literal[0],
    lazy: bool = ...,
    help: str = ...,
) -> NoReturn: ...
@overload
def keyword(
    match: list[str],
    *,
    default: list[T],
    format: Callable[[str], T] = ...,
    num: int,
    lazy: bool = ...,
    help: str = ...,
) -> list[T]: ...
@overload
def keyword(
    match: list[str],
    *,
    default: Optional[T] = ...,
    format: Callable[[str], T],
    num: None = ...,
    lazy: bool = ...,
    help: str = ...,
) -> T: ...
@overload
def keyword(
    match: list[str],
    *,
    default: Optional[list[T]] = ...,
    format: Callable[[str], T],
    num: int,
    lazy: bool = ...,
    help: str = ...,
) -> list[T]: ...
@overload
def keyword(
    match: list[str],
    *,
    default: Any = ...,
    format: Callable[[str], Any] = ...,
    num: int = ...,
    lazy: bool = ...,
    help: str = ...,
    id: str,
) -> ParamSet[Any]: ...

class HelpRequest(Exception): ...

def help_parser() -> Parser[bool]: ...



from __future__ import absolute_import, annotations

import json
import os
import site
from pathlib import Path
from unittest.mock import call, mock_open

from pyfakefs.fake_filesystem import FakeFilesystem
from pytest import fixture
from pytest_mock import MockerFixture

import kslurm.installer.utils as utils


@fixture
def venv(mocker: MockerFixture, fs: FakeFilesystem):
    dir = Path("/fake/path") / "temp-library"

    bin_dir = dir / "venv" / "bin"
    fs.create_dir(bin_dir)
    fs.create_file(dir / "VERSION")

    p_exec = bin_dir / "python"
    fs.create_file(p_exec)
    mocker.patch("sys.executable", p_exec)

    return dir


@fixture
def envvar_dir(mocker: MockerFixture, fs: FakeFilesystem):
    dir = "fake/path/venv"
    fs.create_dir(dir)
    mocker.patch("os.getenv", return_value=dir)
    return Path(dir)


class TestDataDir:
    HOME_DIR = "TESTING_DATA_DIRECTORY"

    def test_returns_current_venv_if_python_executable_is_in_project_venv(
        self, venv: Path, envvar_dir: Path
    ):
        assert utils.data_dir(self.HOME_DIR) == venv

    def test_returns_home_dir_if_set(self, mocker: MockerFixture, envvar_dir: Path):
        getenv_spy = mocker.spy(os, "getenv")
        assert utils.data_dir(self.HOME_DIR) == envvar_dir
        getenv_spy.assert_called_with(self.HOME_DIR)

    def test_checks_xdg_data_if_nothing_else_set(self, mocker: MockerFixture):
        getenv_spy = mocker.spy(os, "getenv")
        assert utils.data_dir(self.HOME_DIR) == Path.home() / ".local/share/kutils"
        getenv_spy.assert_called_with("XDG_DATA_HOME", Path.home() / ".local/share")


class TestBinDir:
    HOME_DIR = "TESTING_DATA_DIRECTORY"

    def test_returns_home_dir_if_set(self, mocker: MockerFixture, envvar_dir: Path):
        getenv_spy = mocker.spy(os, "getenv")
        assert utils.bin_dir(self.HOME_DIR) == envvar_dir / "bin"
        getenv_spy.assert_called_with(self.HOME_DIR)

    def test_returns_default_otherwise(self):
        assert utils.bin_dir(self.HOME_DIR) == Path(site.getuserbase()) / "bin"


def test_get(mocker: MockerFixture):
    mockopen = mock_open(read_data="Hello World")
    urlopen = mocker.patch("kslurm.installer.utils.closing", mockopen)
    text = utils.get("http://google.com")

    assert call().read() in urlopen.mock_calls
    assert text == "Hello World"


class TestGetCurrentVersion:
    def test_returns_version_file_contents_if_exist(self, fs: FakeFilesystem):
        datadir = Path("/fake/datadir")
        fs.create_dir(datadir)
        fs.create_file(datadir / "VERSION", contents="test string")

        assert utils.get_current_version(datadir) == "test string"

    def test_returns_none_if_no_version_file(self, fs: FakeFilesystem):
        datadir = Path("/fake/path")
        fs.create_dir(datadir)

        assert utils.get_current_version(datadir) is None


class TestGetVersion:
    def _mock_data(self, mocker: MockerFixture, add_version: bool = False):
        release_file = Path(__file__).parent / "fake_releases.json"
        with release_file.open() as file:
            data = json.load(file)
        if add_version:
            data["releases"]["1.2.0"] = ""
        mocker.patch("json.loads", return_value=data)
        mocker.patch.object(utils, "get")
        self.fake_url = "https://example.org"

    def test_returns_latest_non_prerelease_normally(self, mocker: MockerFixture):
        self._mock_data(mocker)
        assert utils.get_version(None, False, self.fake_url) == "1.1.10"

    def test_returns_latest_release_when_preview(self, mocker: MockerFixture):
        self._mock_data(mocker)
        assert utils.get_version(None, True, self.fake_url) == "1.2.0a2"

    def test_returns_specified_version_when_given(self, mocker: MockerFixture):
        self._mock_data(mocker)
        assert utils.get_version("1.1.1", False, self.fake_url) == "1.1.1"
        assert utils.get_version("1.0.0-b6", True, self.fake_url) == "1.0.0b6"
        assert utils.get_version("0.8.1.a0", False, self.fake_url) == "0.8.1a0"

    def test_returns_none_when_given_version_doesnt_exist(self, mocker: MockerFixture):
        self._mock_data(mocker)
        assert utils.get_version("1.2.0", True, self.fake_url) is None

    def test_returns_latest_non_dev_version_if_available(self, mocker: MockerFixture):
        self._mock_data(mocker, True)
        assert utils.get_version("1.2.0", True, self.fake_url) == "1.2.0"



from __future__ import absolute_import, annotations

import importlib.resources as impr
import os
import shlex
import sys
from pathlib import Path
from typing import List, Union

import kslurm.appconfig as appconfig
import kslurm.bin
import kslurm.models.job_templates as templates
import kslurm.slurm.helpers as helpers
from kslurm.args.command import Parsers
from kslurm.exceptions import TemplateError, ValidationError
from kslurm.models.slurm import SlurmModel

_LOAD_ENV = (
    "[[ -e $HOME/.bashrc ]] && source $HOME/.bashrc; "
    "source {path}; kpy {method} {name};"
)


class SlurmCommand:
    def __init__(
        self,
        args: Union[SlurmModel, TemplateError],
        tail: list[str],
        arglist: Parsers,
    ):
        self._name = ""
        self._output = ""

        if isinstance(args, TemplateError):
            print(args.msg)
            print("Choose from the following list of templates:\n")
            templates.list_templates()
            exit(1)

        assert isinstance(args, SlurmModel)
        if args.list_job_templates:
            templates.list_templates()
            exit()

        # set_template returns templated values only if a template is passed
        # if we pass a blank string, the models are returned unchanged
        template_vals = templates.set_template(
            args.job_template, mem=args.mem, cpu=args.cpu, time=args.time
        )

        # Start by setting these three with model/template
        self.time = template_vals.time
        self.cpu = template_vals.cpu
        self.mem = template_vals.mem

        # Then update if values were specifically supplied on the command line
        if arglist["time"].value is not None:
            self.time = args.time
        if arglist["cpu"].value is not None:
            self.cpu = args.cpu
        if arglist["mem"].value is not None:
            self.mem = args.mem

        self.gpu = bool(args.gpu)
        self.x11 = bool(args.x11)
        config = appconfig.Config()
        if not args.account:
            if self.gpu:
                account = config.get("account.gpu")
            else:
                account = config.get("account.cpu")
            if account is None:
                account = config.get("account")
            if not account:
                print(
                    "Account must either be specified using --account, or provided in "
                    "config. A default account can be added to config by running "
                    "kslurm config account <account_name>"
                )
                sys.exit(1)
            self.account = account
        else:
            self.account = args.account
        self.cwd = args.directory
        self.test = args.test
        self.job_template = args.job_template
        self._command = tail

        os.chdir(self.cwd)

        self.args = args
        self._venv = args.venv or None

    ###
    # Job Paramaters
    ###
    @property
    def time(self):
        return helpers.slurm_time_format(self._time)

    @time.setter
    def time(self, time: int):
        self._time = time

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, output: str):
        self._output = f'--output="{output}"'

    def set_venv(self, name: str):
        self._venv = name

    ###
    # Command line strings
    ###
    @property
    def slurm_args(self):
        s = (
            f"--account={self.account} --time={self.time} "
            f"--cpus-per-task={self.cpu} --mem={self.mem}"
        )
        if self.gpu:
            s += " --gres=gpu:1"
        if self.x11:
            s += " --x11"
        if self.name:
            s += f" --job-name='{self.name}'"
        return s

    @property
    def command(self):
        return " ".join(self._command)

    @command.setter
    def command(self, commands: List[str]):
        self._command = commands

    @property
    def script(self):
        return "\n".join(["#!/bin/bash", self.venv_load, self.command])

    @property
    def venv_load(self):
        if self._venv:
            with impr.path(kslurm.bin, "kpy-wrapper.sh") as path:
                return _LOAD_ENV.format(path=path, method="load", name=self._venv)
        return ""

    @property
    def venv_activate(self):
        if self._venv:
            with impr.path(kslurm.bin, "kpy-wrapper.sh") as path:
                return _LOAD_ENV.format(path=path, method="activate", name=self._venv)
        return ""

    ###
    # Job submission commands
    ###
    @property
    def run(self):
        if self.command:
            if os.environ.get("SLURM_JOB_ID") or os.environ.get("SLURM_JOBID"):
                return f"srun {self.slurm_args} {self.command}"
            command = self.venv_load + self.command
            return f"srun {self.slurm_args} bash -c {shlex.quote(command)}"
        command = (
            "srun --pty bash -c "
            + shlex.quote(f'bash --init-file <(echo "{self.venv_load}";)')
            if self.venv_load
            else ""
        )
        return f"salloc {self.slurm_args} {command}"

    @property
    def batch(self):
        if self.test:
            s = "cat"
        else:
            s = f"sbatch {self.slurm_args} --parsable {self.output}"
        if self.command:
            if "/" in self._command[0] and Path(self._command[0]).is_file():
                try:
                    with open(self._command[0], "r", encoding="utf-8") as f:
                        if f.readline().startswith("#/"):
                            return f"{s} {self.command}"
                except UnicodeDecodeError:
                    pass
            return f"echo {shlex.quote(self.script)} | {s}"
        else:
            raise ValidationError("No command given")

    @property
    def batch_test(self):
        s = f"sbatch {self.slurm_args} --test-only"
        if self.command:
            return f"echo {shlex.quote(self.script)} | {s}"
        else:
            raise ValidationError("No command given")



from __future__ import absolute_import, annotations

import itertools as it
import operator as op
from pathlib import Path
from typing import Any, DefaultDict, Optional, Union

import attrs
from rich.console import Group
from rich.padding import Padding
from rich.table import Table
from rich.text import Text
from typing_extensions import Self

from kslurm.args.arg import SKIPHELP, AbstractHelpTemplate, Arg, ParamSet


class BasicTemplate(AbstractHelpTemplate):
    title = "options"
    header = []
    cls_usage = "<options>"
    priority = 50

    def row(self, name: str, help: str, default: Optional[str]):
        return (
            [name, help, default if default is not None else ""]
            if help is not SKIPHELP
            else None
        )

    def update_meta(self, **kwargs: Any) -> Self:
        raise NotImplementedError()

    @property
    def help(self):
        return ""


def _header(header: str):
    return f"[bold green]{header.upper()}[/bold green]"


def _section(header: str, description: str | None, body: Union[Table, str]):
    return Group(
        *filter(
            None,
            [
                Padding.indent(_header(header), 4),
                Padding.indent(description, 8) if description else None,
                Padding.indent(body, 8) if body else None,
                "\n",
            ],
        )
    )


@attrs.frozen
class HelpText:
    script: str
    models: dict[str, Union[Arg[Any], ParamSet[Any]]]
    script_help: str = ""
    usage_suffix: str = ""
    just_usage: bool = False

    def with_usage_only(self):
        return attrs.evolve(self, just_usage=True)

    def __str__(self):
        return str(self.__rich__())

    def __rich__(self):
        models = self.models
        script_name = Path(self.script).name

        templates: set[type[AbstractHelpTemplate]] = set()
        usages: dict[type[AbstractHelpTemplate], list[str]] = DefaultDict(list)
        for model in models.values():
            if model.help_template is None:
                template = BasicTemplate()
            else:
                template = model.help_template
            args = model.label, model.help, str(model.assigned_value or "")
            template.add_row(*args)
            try:
                usages[type(template)].append(template.usage(*args))
            except NotImplementedError:
                pass
            templates.add(type(template))

        sorted_templates: dict[type[AbstractHelpTemplate], list[str]] = {}

        for template in sorted(usages, key=op.attrgetter("priority"), reverse=True):
            if not usages[template]:
                usages[template].append(template.cls_usage)
            sorted_templates[template] = usages[template]

        command_line_example = Text.assemble(
            ("USAGE: ", "b"),
            script_name,
            Text.from_markup(
                " " + " ".join(it.chain.from_iterable(sorted_templates.values()))
            )
            if usages
            else "",
            (f" <{self.usage_suffix}>", "cyan") if self.usage_suffix else "",
            "\n",
        )
        if self.just_usage:
            return Group(command_line_example)

        return Group(
            *filter(
                None,
                [
                    command_line_example,
                    self.script_help + "\n" if self.script_help else "",
                    *(
                        _section(template.title, template.description, template.table())
                        for template in templates
                    ),
                ],
            )
        )



#!/usr/bin/env python3
import os
import sys

from snakemake.utils import read_job_properties

# last command-line argument is the job script
jobscript = sys.argv[-1]

# all other command-line arguments are the dependencies
dependencies = set(sys.argv[1:-1])


# parse the job script for the job properties that are encoded by snakemake within
job_properties = read_job_properties(jobscript)


# get account from CC_COMPUTE_ALLOC, else supply default account
account = os.getenv('CC_COMPUTE_ALLOC', default='ctb-akhanf')

   

if job_properties["type"]=='single':
    
    #this is a single rule, so set the resource request based on the rule itself
    # or to single rule defaults (1 core, 1hr, 4gb)
    job_name = job_properties['rule']

    #get values and set defaults
    if 'time' in job_properties["resources"].keys():
        time = job_properties["resources"]["time"]
    else:  
        time = 60

    if 'mem_mb' in job_properties["resources"].keys():
        mem_mb = job_properties["resources"]["mem_mb"]
    else:  
        mem_mb = 4000

    if 'gpus' in job_properties["resources"].keys():
        gpus = job_properties["resources"]["gpus"]
    else:  
        gpus = 0

    #for single job, threads already set to 1 by default
    threads = job_properties["threads"]


elif job_properties["type"]=='group':

    #for a grouped rule, we request an entire node in order to maximize throughput:
    job_name = job_properties['groupid']

    # Maybe look into later: add some checks to 
    #  1) set a reasonable walltime (less than 24hrs) based on 1hr 1core jobs
    #  2) check when there may be too many jobs in the group
    # if 2), could always try to resubmit with larger job?  
    
    time = 60*24
    mem_mb = 128000
    threads = 32
    gpus = 0

else:
    raise NotImplementedError(f"Don't know what to do with job_properties['type']=={job_properties['type']}")



log = os.path.realpath(os.path.join('logs','slurm',f'slurm_%j_{job_name}.out'))

#create the log directory (slurm fails if doesn't exist)
log_dir = os.path.dirname(log)
if not os.path.exists(os.path.dirname(log)):
    os.makedirs(log_dir)



# collect all command-line options in an array
cmdline = ["sbatch"]

if gpus > 0:
    gpu_arg = f'--gres=gpu:t4:{gpus}'
else:
    gpu_arg = ''

# set all the slurm submit options as before
slurm_args = f" --parsable --account={account} {gpu_arg} --time={time} --mem={mem_mb} --cpus-per-task={threads} --output={log} "

cmdline.append(slurm_args)

if dependencies:
    cmdline.append("--dependency")
    # only keep numbers in dependencies list
    dependencies = [ x for x in dependencies if x.isdigit() ]
    cmdline.append("afterok:" + ",".join(dependencies))

cmdline.append(jobscript)

os.system(" ".join(cmdline))



from __future__ import absolute_import

import abc
import re
from pathlib import Path
from typing import Any, Callable, Literal, Optional, overload

import attrs

from kslurm.args.arg import Context, Parser


class BasicMatcher(abc.ABC):
    duplicates: bool = False
    max_len: Optional[int] = None

    def __call__(self, arg: str, param: Parser[Any], context: Context):
        if param.value is None and param.validation_err is None:
            return self.test(arg)
        if not self.duplicates:
            return False
        if self.max_len is not None:
            try:
                if len(param.value or "") > self.max_len:
                    return False
            except KeyError:
                pass
        return self.test(arg)

    @abc.abstractmethod
    def test(self, arg: str) -> bool:
        raise NotImplementedError()

    def settings(self, duplicates: bool = False, max_len: Optional[int] = None):
        self.duplicates = duplicates
        self.max_len = max_len
        return self


class choice(BasicMatcher):
    def __init__(self, *choices: str):
        self.choices = choices

    def test(self, arg: str):
        return arg in self.choices


@attrs.frozen
class regex(BasicMatcher):
    pattern: str

    def test(self, arg: str):
        return bool(re.match(self.pattern, arg))


@attrs.frozen
class path(BasicMatcher):
    is_dir: bool = False

    def test(self, arg: str):
        is_path = "/" in arg and Path(arg).exists()
        if self.is_dir:
            return is_path and Path(arg).is_dir()
        return is_path


class everything(BasicMatcher):
    def test(self, arg: str):
        return True


@overload
def option_chain(
    initializer: str, number: Literal[1], matcher: BasicMatcher
) -> Callable[[str, Parser[Any], Context], bool]:
    ...


@overload
def option_chain(
    initializer: str, number: Optional[int], matcher: BasicMatcher
) -> Callable[[str, Parser[list[Any]], Context], bool]:
    ...


def option_chain(initializer: str, number: Optional[int], matcher: BasicMatcher) -> Any:
    def inner_single(arg: str, _: Parser[Any], context: Context):
        if context.last_matched is not None and context.last_matched.id == initializer:
            return matcher.test(arg)
        return False

    def inner_chain(arg: str, param: Parser[list[Any]], context: Context):
        if (
            context.last_matched is not None
            and context.last_matched.id in [initializer, param.id]
            and (number is None or len(param.value or []) <= number)
        ):
            return matcher.test(arg)
        return False

    if number == 1:
        return inner_single
    else:
        return inner_chain



from __future__ import absolute_import, annotations

import itertools as it
from typing import Any, Optional, Union

import attr
import docstring_parser as doc
import more_itertools as itx
from rich.text import Text
from tabulate import tabulate
from typing_extensions import Self

from kslurm.args.arg import AbstractHelpTemplate, HelpRow
from kslurm.args.protocols import WrappedCommand


@attr.frozen
class ShapeArg(AbstractHelpTemplate):
    title = "Shape Args"
    header = ["", "Syntax", "", "Examples", "Default", ""]
    right_align_cols = 2

    syntax: str
    examples: list[str]

    def _syntax_format(self, syntax: str):
        return Text.assemble(
            *itx.interleave(
                [Text(line.strip(), style="cyan bold") for line in syntax.split(" | ")],
                it.repeat("\n"),
            )
        )

    def update_meta(self, **kwargs: Any) -> Self:
        return attr.evolve(self, **kwargs)

    def usage(self, name: str, help: str, default: Optional[str]) -> str:
        return rf"[hot]\[{name}][/]"

    def row(self, name: str, help: str, default: Optional[str]):
        return [
            Text(name, style="bold"),
            self._syntax_format(self.syntax),
            ("➔" if self.examples else ""),
            Text("\n".join(itx.always_iterable(self.examples))),
            Text(str(default), style="default_col") if default is not None else "",
            Text(help),
        ]

    @property
    def help(self):
        return ""


@attr.frozen
class PositionalArg(AbstractHelpTemplate):
    title = "Positional Args"
    header = []

    def usage(self, name: str, help: str, default: Optional[str]) -> str:
        if default is None:
            return f"[cyan]{name}[/]"
        return rf"[cyan]\[{name}][/]"

    def update_meta(self, **kwargs: Any) -> Self:
        raise NotImplementedError()

    def row(self, name: str, help: str, default: Optional[str]):
        return [
            Text(name, style="bold"),
            Text(default if default is not None else "REQUIRED", style="default_col"),
            Text(help),
        ]

    @property
    def help(self):
        return ""


@attr.frozen
class SubcommandTemplate(AbstractHelpTemplate):
    title = "Commands"
    description = "Run any command followed by -h for more information"
    header = []

    commands: dict[str, WrappedCommand]

    def usage(self, name: str, help: str, default: Optional[str]) -> str:
        return f"[cyan]{name}[/]"

    def update_meta(self, **kwargs: Any) -> Self:
        raise NotImplementedError()

    def row(
        self, name: str, help: str, default: Optional[str]
    ) -> Union[list[HelpRow], HelpRow]:
        return [
            [],
            *self.command_list,
        ]

    @property
    def command_list(self):
        return (
            [
                name,
                doc.parse(func.__doc__ or "").short_description or "",
            ]
            for name, func in self.commands.items()
        )

    @property
    def help(self):
        return tabulate(list(self.command_list), tablefmt="plain")
        # t = Table.grid(padding=(0, 2), expand=True)
        # for _ in range(2):
        #     t.add_column(justify="right", no_wrap=True)
        # for r in self.command_list:
        #     t.add_row(*r)
        # return t


# @attr.frozen
# class KeywordArg(AbstractHelpEntry):
#     title = "Keyword Args"
#     header = ["", "Default", ""]

#     name: list[str]
#     value_name: Optional[str]
#     default: str
#     help: str

#     @property
#     def usage(self):
#         return ""

#     def row(self) -> list[Union[Text, str]]:
#         return [
#             Text(", ".join(self.name)) + (
#                 Text(f" <{(self.value_name)}>", style="grey")
#                 if self.value_name else ''
#             ),
#             Text(self.default, style="default_col"),
#             Text(textwrap.fill(self.help, 70)),
#         ]


# @attr.frozen
# class ChoiceArg(AbstractHelpEntry):
#     title = "Choice Args"
#     header = ["", "Choices", "Default", ""]

#     name: str
#     choices: list[str]
#     default: str
#     help: str

#     @property
#     def usage(self):
#         return ""

#     def row(self) -> list[Union[Text, str]]:
#         return [
#             Text(self.name, style="bold"),
#             Text(",".join(self.choices), style="bold"),
#             Text(self.default, style="default_col"),
#             Text(textwrap.fill(self.help, 70)),
#         ]



from __future__ import absolute_import, annotations

import functools as ft
import inspect
import itertools as it
import sys
from typing import Any, Callable, Literal, Optional, TypeVar, Union, overload

import attr
import docstring_parser as doc
from rich.text import Text
from typing_extensions import ParamSpec

from kslurm.args.arg import Arg, Parser
from kslurm.args.arg_types import HelpRequest, help_parser
from kslurm.args.help import HelpText
from kslurm.args.helpers import get_arg_dict, get_parsers, read_parsers
from kslurm.args.parser import parse_args
from kslurm.args.protocols import Command
from kslurm.exceptions import CommandLineError, TailError
from kslurm.style import console, stderr

P = ParamSpec("P")
_CommandFunc = Callable[P, Optional[int]]
ModelType = Union[dict[str, Arg[Any]], type]


Parsers = dict[str, Parser[Any]]


def error(argv: list[str] = []):
    return 1


class CommandArgs:
    _model: Optional[str] = None
    _tail: Optional[str] = None
    _modellist: Optional[str] = None
    _name: Optional[str] = None
    _helptext: Optional[str] = None

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value: str):
        if self._model is not None:
            raise ValueError("model is already set")
        self._model = value

    @property
    def tail(self):
        return self._tail

    @tail.setter
    def tail(self, value: str):
        if self._tail is not None:
            raise ValueError("tail is already set")
        self._tail = value

    @property
    def modellist(self):
        return self._modellist

    @modellist.setter
    def modellist(self, value: str):
        if self._modellist is not None:
            raise ValueError("modellist is already set")
        self._modellist = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value: str):
        if self._name is not None:
            raise ValueError("name is already set")
        self._name = value

    @property
    def helptext(self):
        return self._helptext

    @helptext.setter
    def helptext(self, value: str):
        if self._helptext is not None:
            raise ValueError("helptext is already set")
        self._helptext = value


Exc = TypeVar("Exc", bound=type[Exception], covariant=True)


class CommandError(Exception):
    def __init__(self, msg: str, *args: Any):
        super().__init__(msg, *args)
        self.msg = msg


@overload
def command(
    maybe_func: None = ...,
    *,
    terminate_on_unknown: bool = ...,
    inline: bool = ...,
    allow_unknown: bool = False,
) -> Callable[[_CommandFunc[P]], Command[P]]:
    ...


@overload
def command(
    maybe_func: _CommandFunc[P],
    *,
    terminate_on_unknown: Literal[True] = ...,
    inline: Literal[False] = ...,
    allow_unknown: bool = False,
) -> Command[P]:
    ...


def command(
    maybe_func: Optional[_CommandFunc[P]] = None,
    *,
    terminate_on_unknown: bool = False,
    inline: bool = False,
    allow_unknown: bool = False,
) -> Union[Command[P], Callable[[_CommandFunc[P]], Command[P]]]:
    @attr.frozen
    class BlankModel:
        pass

    def decorator(func: _CommandFunc[P]):
        unknown = allow_unknown
        if not callable(func):
            raise CommandLineError(f"{func} is not callable")

        params = inspect.signature(func)
        exceptions: tuple[type[Exception], ...] = tuple()  # type: ignore
        model = BlankModel

        command_args = CommandArgs()
        if inline:
            model = attr.make_class(
                f"{func.__name__}__Model__",
                {
                    param.name: attr.attrib(
                        default=param.default
                        if param.default is not params.empty
                        else attr.NOTHING,
                        type=param.annotation
                        if param.annotation is not params.empty
                        else attr.NOTHING,
                    )
                    for param in params.parameters.values()
                },
            )
            model.__doc__ = func.__doc__
        else:
            for param in params.parameters.values():
                annotation = (
                    eval(param.annotation, func.__globals__)
                    if isinstance(param.annotation, str)
                    else param.annotation
                )
                if annotation == param.empty:
                    raise CommandLineError(
                        f"parameter '{param}' in {func} must be annotated"
                    )
                if annotation == list[str]:
                    command_args.tail = param.name
                    unknown = True
                    continue
                if annotation == Parsers:
                    command_args.modellist = param.name
                    continue
                if annotation == str:
                    command_args.name = param.name
                    continue
                if annotation == HelpText:
                    command_args.helptext = param.name
                    continue

                if model is not BlankModel:
                    raise CommandLineError(
                        f"Mutliple models detected: {model} and {annotation}"
                    )
                model = annotation

                types = getattr(model, "__args__", None)
                if types is not None:
                    model = types[0]
                    exceptions = types[1:]

                if not attr.has(model):
                    raise CommandLineError(
                        f"Annotation of {param} in {func} must be an attr "
                        f"annotated class (currently {model})"
                    )
                command_args.model = param.name

        docobj = doc.parse(func.__doc__ or "")
        docstr = (
            (docobj.short_description or "")
            + (docobj.blank_after_short_description and "\n\n" or "\n")
            + (docobj.long_description or "")
        )

        model_dict = get_arg_dict(model)
        if command_args.tail:
            usage_suffix = "command_args"
        else:
            usage_suffix = ""

        class Wrapper:
            def get_helptext(self, entrypoint: str):
                return HelpText(entrypoint, model_dict, docstr, usage_suffix)

            @ft.wraps(func)
            def __call__(self, *args: P.args, **kwargs: P.kwargs):
                if inline:
                    for param in it.islice(
                        params.parameters.values(), len(args), len(params.parameters)
                    ):
                        if (
                            isinstance(param.default, Arg)
                            and param.default.assigned_value is None  # type: ignore
                        ):
                            raise SyntaxError(
                                f"Mandatory param '{param}' in function '{func}' not "
                                "provided a value"
                            )
                return func(*args, **kwargs) or 0

            @ft.wraps(func)
            def cli(self, argv: list[str] = sys.argv):
                helptext = self.get_helptext(argv[0])
                try:
                    parsed_list, tail = parse_args(
                        argv[1:],
                        {
                            **get_parsers(model_dict),
                            "help": help_parser().with_id("help"),
                        },
                        terminate_on_unknown=terminate_on_unknown,
                        allow_unknown=unknown,
                    )

                    parsed, errors = read_parsers(model_dict, parsed_list, False, False)

                    if errors or isinstance(tail, TailError):
                        console.print(helptext.with_usage_only())
                        for error in errors.values():
                            stderr.print(Text.from_ansi(error.msg))
                        if isinstance(tail, TailError):
                            print(tail.msg, file=sys.stderr)
                        return 1
                    if not inline:
                        args: dict[str, Any] = {}

                        if command_args.model is not None:
                            args[command_args.model] = model(**parsed)

                        if command_args.tail is not None:
                            args[command_args.tail] = tail

                        if command_args.modellist is not None:
                            args[command_args.modellist] = parsed_list

                        if command_args.name is not None:
                            args[command_args.name] = argv[0]

                        if command_args.helptext is not None:
                            args[command_args.helptext] = helptext
                    else:
                        args = parsed

                except (HelpRequest, *exceptions) as err:
                    if isinstance(err, HelpRequest) and not isinstance(err, exceptions):
                        console.print(helptext)
                        return 0

                    args: dict[str, Any] = {}
                    if command_args.model is not None:
                        args[command_args.model] = err

                    if command_args.tail is not None:
                        args[command_args.tail] = argv[1:]

                    if command_args.modellist is not None:
                        args[command_args.modellist] = {}

                    if command_args.name is not None:
                        args[command_args.name] = argv[0]

                    if command_args.helptext is not None:
                        args[command_args.helptext] = helptext
                try:
                    return func(**args) or 0  # type: ignore
                except CommandError as err:
                    print(err.msg, file=sys.stderr)
                    return 1

        return Wrapper()

    if maybe_func is None:
        return decorator
    return decorator(maybe_func)



from __future__ import absolute_import, annotations

SETTINGS_HEADER = "[sky_blue1]SETTINGS[/]"
COMMAND_HEADER = "[sky_blue1]COMMAND[/]"

KBATCH_MSG = f"""
    [green]Scheduling Batch Command[/]
        {SETTINGS_HEADER}
            [white]{{slurm_args}}[/]
        {COMMAND_HEADER}
            {{command}}
"""

KRUN_CMD_MESSAGE = f""" \
    [green]Running job[/]
        {SETTINGS_HEADER}
            [white]{{args}}[/]
        {COMMAND_HEADER}
            [white]{{command}}[/]
"""

INTERACTIVE_MSG = f"""
    [green]Running interactive session[/]
        {SETTINGS_HEADER}
            [white]{{args}}[/]
"""

JUPYTER_WELCOME = """
    [hot]Started Jupyter Server!![/]

    [heading]Web browser access:[/]
        You'll need to set up an ssh tunnel. Open a new console
        and enter the tunnel script below.
        Then paste the browser url into your browser of choice!

    [heading]VS Code access:[/]
        Start a VS Code SSH session (see here for more info:
        https://code.visualstudio.com/docs/remote/ssh). Open your
        Jupyter notebook file, then set the Jupyter server to remote:
            Command Pallete > "Jupyter: Specify local or remote Jupyter server for
            connections"
        When prompted, enter the Server URL

    [heading]Tunnel script (replace <hostname> with your own hostname):[/]
        {tunnel_script}
    [heading]Browser URL:[/]
        {browser_url}
    [heading]Server URL:[/]
        {url}

    Once in the interactive shell, you can use kjupyter subcommands to interact with
    jupyter:
        - kjupyter log      View the server logs
        - kjupyter console  Start an interactive ipython console
        - kjupyter url      View the server url
        - kjupyter tunnel   Echo bash code to create an ssh tunnel to the server
"""



# pyright: reportPrivateUsage=false, reportUnknownArgumentType=false
from __future__ import absolute_import

from copy import copy
from typing import Any, NamedTuple

import hypothesis.strategies as st
from hypothesis import assume, given

import kslurm.args.parser as sc
from kslurm.args import actions, helpers, keyword, matchers
from kslurm.args.arg import Context, ParamInterface, Parser


class CliParsers(NamedTuple):
    argv: list[str]
    models: dict[str, Parser[Any]]
    results: dict[str, Parser[Any]]
    tail: list[str]


class CliParams(NamedTuple):
    argv: list[str]
    models: ParamInterface[Any]
    results: Any


class CliParamsZipped(NamedTuple):
    argv: tuple[list[str]]
    models: tuple[dict[str, ParamInterface[Any]]]
    results: tuple[dict[str, Any]]


@st.composite
def keyword_arg(draw: st.DrawFn, match_space: list[str] = []):
    alphabet = st.characters(blacklist_categories=("Cs", "Zs"))
    if not match_space:
        match = draw(st.text(alphabet, min_size=1))
    else:
        match = draw(st.sampled_from(match_space))
    entries = draw(
        st.lists(
            st.text(alphabet).filter(lambda s: s not in match_space),
            min_size=1,
            max_size=10,
        )
    )
    size = len(entries)
    argv = [match, *entries]
    models: ParamInterface[Any] = keyword(
        match=[match], default=None, num=size, id=match
    )

    return CliParams(argv, models, entries)


def arg():
    return st.text(st.characters(blacklist_categories=("Cs", "Zs")), min_size=1)


def match_space():
    return st.lists(arg(), min_size=5, max_size=20, unique=True)


@st.composite
def cli_params(draw: st.DrawFn):
    mspace = draw(match_space())

    priorities = draw(
        st.lists(
            st.integers(min_value=0, max_value=20), min_size=0, max_size=8, unique=True
        ),
    )

    matched: set[str] = set()
    models: dict[str, Parser[Any]] = {}
    argv: list[str] = []
    results = copy(models)
    tail: list[str] = []
    for i, p in enumerate(sorted(priorities, reverse=True)):
        matches = draw(
            st.lists(
                st.sampled_from(mspace),
                min_size=0,
                max_size=len(mspace),
                unique=True,
            )
        )

        id = f"{i}/{p} - ({matches})"
        models[id] = Parser(
            id=id,
            priority=p,
            match=matchers.choice(*matches),
            action=actions.replace(),
        )
        results[id] = models[id]
        unmatched = set(matches) - matched
        if unmatched:
            candidate = draw(st.sampled_from(list(unmatched)))
            argv.append(candidate)
            results[id] = models[id].with_value(candidate, Context([], 0, {}, None))
            matched.update(matches)

            # Add extra tail args
            for _ in range(draw(st.integers(min_value=0, max_value=2))):
                extra = draw(arg())
                assume(extra not in mspace)
                argv.append(extra)
                tail.append(extra)

    return CliParsers(argv, models, results, tail)


# @st.composite
# def keywords_and_flags(draw: st.DrawFn):
#     mspace = draw(match_space())
#     kwarg = draw(st.lists(keyword_arg()))
#     if len(kwarg) > 1:
#         argv, kw, results = cast(CliParamsZipped, zip(*kwarg))
#     return
#     number = draw(st.integers(min_value=0, max_value=5))
#     matches: list[str] = []
#     for _ in range(number):
#         match = draw(st.text(alphabet=alphabet))
#         assume(match not in matches)
#         assume(match)
#         matches.append(match)

#     num_args = [
#         draw(st.one_of(st.just(0), st.integers(min_value=0, max_value=5)))
#         for _ in range(number)
#     ]
#     priorities = [draw(st.integers(min_value=0, max_value=20)) for _ in range(number)]
#   greedinesses = [draw(st.integers(min_value=0, max_value=20)) for _ in range(number)]

#     model_list: list[Arg[Any, Any]] = []
#     for i in range(len(matches)):
#         id = f"{matches[i]}/{priorities[i]}"
#         model = Arg(
#             id=id,
#             priority=priorities[i],
#             match=matchers.choice(matches[i]),
#             format=lambda arg, *_: True,  # type: ignore
#             default=False,
#         )
#         model_list.append(model)

#     models: dict[str, Arg[Any, Any]] = {}
#     for i in range(len(matches)):
#         id = model_list[i].id
#         models[id] = model_list[i]
#         if num_args[i]:
#             models[id + " (args)"] = Arg(
#                 id=id + " (args)",
#                 priority=greedinesses[i],
#                 match=matchers.option_chain(id, num_args[i], matchers.everything()),
#                 format=lambda arg, *_: arg,  # type: ignore
#             )

#     argv: list[str] = []
#     tail: list[str] = []
#     results = copy(models)
#     prev_priority = 0
#     for _ in range(draw(st.integers(min_value=0, max_value=number)) * 2):
#         higher_priority = [
#             match for i, match in enumerate(matches) if priorities[i] >= prev_priority
#         ]
#         if not len(higher_priority):
#             break
#         index = matches.index(draw(st.sampled_from(higher_priority)))
#         argv.append(matches[index])
#         num_params = draw(st.integers(min_value=0, max_value=num_args[index]))
#         params: list[str] = []
#         for _ in range(num_params):
#             lower_priority = [
#                 match
#                 for i, match in enumerate(matches)
#                 if priorities[i] < greedinesses[index]
#             ]
#             higher_priority = [
#                 match for match in matches if match not in lower_priority
#             ]
#             if len(lower_priority):
#                 param = draw(
#                     st.one_of(
#                         st.text(alphabet=alphabet), st.sampled_from(lower_priority)
#                     )
#                 )
#             else:
#                 param = draw(st.text(alphabet=alphabet))
#             assume(param not in higher_priority)
#             params.append(param)
#         prev_priority = greedinesses[index]
#         argv.extend(params)
#         id = model_list[index].id
#         results[id] = attrs.evolve(models[id], value=True)
#         results[id + " (args)"] = attrs.evolve(results[id + " (args)"], value=params)

#         if num_params == num_args[index]:
#             extra_args = draw(st.integers(min_value=0, max_value=2))
#             for _ in range(extra_args):
#                 extra = draw(st.text(alphabet=alphabet))
#                 assume(extra not in matches)
#                 argv.append(extra)
#                 tail.append(extra)

#     return CliParams(argv, models, results, tail)


@given(keyword_arg())
def test_keyword_arg(params: CliParams):
    argv, models, template = params
    parsers = helpers.get_parsers({models.id: models})
    parsed, tail = sc.parse_args(argv, parsers)
    print(parsed)
    results, err = helpers.read_parsers({models.id: models}, parsed)
    assert err == {}
    assert results == {models.id: template}
    assert tail == []


@given(cli_params())
def test_parser(params: CliParsers):
    argv, models, results, tail = params
    a_result, a_tail = sc._match_args(argv, models, terminate_on_unknown=False)
    assert results == a_result
    assert tail == a_tail



from __future__ import absolute_import, annotations

import re

from kslurm.exceptions import TemplateError, ValidationError
from kslurm.models.job_templates import templates


def job_template(arg: str) -> str:
    if arg in templates():
        return arg
    raise TemplateError(f"{arg} is not a valid job-template")


_FS_NAME_PATTERN = re.compile(r"^[\w\-\. ]+$")


def fs_name(name: str):
    if name and not re.match(_FS_NAME_PATTERN, name):
        raise ValidationError(f"Invalid characters found in {name}")
    return name



from __future__ import absolute_import, annotations

import re

import pint

from kslurm.exceptions import ValidationError

ureg = pint.UnitRegistry()


def time(time: str):
    if ":" in time:
        if "-" in time:
            days, hhmm = time.split("-")
            hours, min = hhmm.split(":")
        else:
            days = "0"
            hours, min = time.split(":")
    else:
        raise TypeError(
            f'Invalid format for time: "{time}"\n'
            f"Must be as [xx-]xx:xx or x where x is a number"
        )
    return int(min) + int(hours) * 60 + int(days) * 60 * 24


def mem(mem: str):
    if not (match := re.match(r"^([0-9]+[GMKgmk][Ii]?)[bB]?$", mem)):
        raise ValidationError(
            "Memory is not formatted correctly. Must be xxx(G|M)[B], e.g. 32G, 4000MB, "
            "etc"
        )
    return int(
        ureg(
            (match.group(1) + "B")
            .upper()
            .replace("K", "k")
            .replace("I", "i")
            .replace("ki", "Ki")
        )
        .to("megabyte")  # type: ignore
        .m
    )



from __future__ import absolute_import

import datetime as dt
import re
import subprocess
from typing import Union

from colorama import Fore

import kslurm.text as txt
from kslurm.args.command import Parsers, command
from kslurm.exceptions import TemplateError
from kslurm.models.slurm import SlurmModel
from kslurm.slurm.slurm_command import SlurmCommand
from kslurm.style import console


@command(terminate_on_unknown=True)
def kbatch(
    args: Union[SlurmModel, TemplateError],
    command_args: list[str],
    arglist: Parsers,
):
    """Submit a job using sbatch

    Supports scripts (e.g. ./script.sh) or direct commands (e.g. cp dir/file.txt dir2)

    When your command contains bash interpreted elements such as $VARIABLES and
    $(subshells), these will be immediately expanded. Normally, this behaviour
    is fine, but sometimes they should only be interpretted on the allocated cluster.
    For instance, $SLURM_TMPDIR will evaluate to "" unless it is interpretted later.
    To force this behaviour, wrap the $VARIABLE or $(subshell) in quotes:
        '$SLURM_TMPDIR'
        '$(hostname)'
    """

    slurm = SlurmCommand(args, command_args, arglist)
    command = slurm.command if slurm.command else f"{Fore.RED}Must provide a command"

    console.print(txt.KBATCH_MSG.format(slurm_args=slurm.slurm_args, command=command))
    if slurm.command:
        test = subprocess.run(slurm.batch_test, shell=True, capture_output=True)

        proc = subprocess.run(
            slurm.batch, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )

        out = proc.stdout.decode()

        if proc.returncode != 0:
            print(Fore.WHITE + out)
            return 1
        if slurm.test:
            # output will be the issued command, so we print it
            print(Fore.WHITE + out)

        if test.returncode == 0 and (
            match := re.search(
                r"(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})T"
                r"(?P<hour>\d{2}):(?P<minute>\d{2}):(?P<second>\d{2})",
                test.stderr.decode(),
            )
        ):
            startdate = dt.datetime(
                **{key: int(value) for key, value in match.groupdict().items()}
            )
            now = dt.datetime.now()
            delta = startdate - now

            prefix = "at"
            # If at least 7am tomorrow
            if delta.total_seconds() > 16 * 60 * 60 and startdate.date() != now.date():
                if startdate.date() == now.date() + dt.timedelta(days=1):
                    date = " tomorrow"
                elif startdate.date() < now.date() + dt.timedelta(days=6):
                    date = f" on {startdate.strftime('%A')}"
                else:
                    date = f" on {startdate.strftime('%a, %b %d')}"
            else:
                date = ""
            time = startdate.strftime("%I:%M%p")

            ddhr, mn = divmod(int(max(0, delta.total_seconds())) // 60, 60)
            day, hr = divmod(ddhr, 24)
            _fromnow = ",".join(
                filter(
                    None,
                    [
                        f"{day}d" if day else None,
                        f"{hr}h" if day or hr else None,
                        f"{mn}m" if day or hr or mn else None,
                    ],
                )
            )
            fromnow = f"({_fromnow} from now)" if _fromnow else "(immediately)"
        else:
            prefix = ""
            date = ""
            time = ""
            fromnow = ""

        if slurm.test:
            print(
                f"""Estimated start time {prefix} {Fore.BLUE}{time}{date} \
{Fore.LIGHTRED_EX}{fromnow}
"""
            )
            return
        slurmid = out.strip()
        print(
            f"""Scheduled job {Fore.LIGHTBLACK_EX}{slurmid}
    {Fore.WHITE}Estimated start {prefix} {Fore.BLUE}{time}{date} \
{Fore.LIGHTRED_EX}{fromnow}
    {Fore.WHITE}To cancel, run:
        scancel {slurmid}
        """
        )


if __name__ == "__main__":
    kbatch.cli(["kbatch", "--account", "foo", "-t"])



from __future__ import absolute_import, annotations

import functools as ft
import importlib.resources as impr
import itertools as it
import json
from typing import Dict, cast

import attr
from tabulate import tabulate
from typing_extensions import TypedDict

import kslurm.models.formatters as formatters


class Templates(TypedDict):
    cpus: str
    mem: str
    time: str


@attr.s(auto_attribs=True)
class TemplateArgs:
    mem: int
    cpu: int
    time: int


@ft.cache
def templates() -> Dict[str, Templates]:
    with impr.open_text("kslurm.data", "slurm_job_templates.json") as stream:
        data = json.load(stream)
    cast(Dict[str, Templates], data)
    return data


def set_template(template: str, mem: int, cpu: int, time: int):
    if template:
        if template not in templates():
            raise Exception(f"{template} is not a valid template")

        return TemplateArgs(
            mem=int(templates()[template]["mem"]),
            cpu=int(templates()[template]["cpus"]),
            time=formatters.time(templates()[template]["time"]),
        )
    else:
        return TemplateArgs(mem=mem, cpu=cpu, time=time)


def list_templates():
    labelled_values = list(templates().values())
    headers = ["name"] + list(labelled_values[0].keys())
    values = [list(value.values()) for value in labelled_values]
    entries = list(zip(templates().keys(), values))
    table = [list(it.chain([entry[0]], entry[1])) for entry in entries]
    print(tabulate(table, headers=headers, tablefmt="presto"))



__submodules__ = ["command", "arg_types"]

__ignore__ = ["T", "S", "C", "P", "Exc"]

# <AUTOGEN_INIT>
from kslurm.args.command import (
    CommandArgs,
    CommandError,
    ModelType,
    Parsers,
    command,
    error,
)
from kslurm.args.arg_types import (
    HelpRequest,
    InvalidSubcommand,
    Subcommand,
    choice,
    flag,
    help_parser,
    keyword,
    path,
    positional,
    shape,
    subcommand,
)

__all__ = [
    "CommandArgs",
    "CommandError",
    "HelpRequest",
    "InvalidSubcommand",
    "ModelType",
    "Parsers",
    "Subcommand",
    "choice",
    "command",
    "error",
    "flag",
    "help_parser",
    "keyword",
    "path",
    "positional",
    "shape",
    "subcommand",
]

# </AUTOGEN_INIT>



from __future__ import absolute_import, annotations

from typing import Generic

from typing_extensions import ParamSpec, Protocol

from kslurm.args.help import HelpText

P = ParamSpec("P")


class WrappedCommand(Protocol):
    def __call__(self, argv: list[str] = ...) -> int:
        ...


class Command(Protocol, Generic[P]):
    def get_helptext(self, entrypoint: str) -> HelpText:
        ...

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> int:
        ...

    def cli(self, argv: list[str] = ...) -> int:
        ...



#!/usr/bin/env python3

# Aim: validate bids-apps.tsv
#    each row: have 4 columns
#    2nd column(url):must star with "docker://" or "shub://", for instance 
#        gradcorrect_v0.0.2', 'docker://khanlab/gradcorrect:v0.0.2
#    1st column(name): match 2nd columns' docker image name and tag part
#    2nd column(url): docker/shub image exist on the hub.
#    3rd column(base_option): '' or start with '-'
#    4th column(job_template): a string match a job name in folder job-templates
#
# Note:
#    use '#' to ignore the validation a line in bids-apps.tsv

import csv
import os

# set validate_image to Ture, will validate docker image(docker pull) and singularity image(singularity pull), may take a very long time...
validate_image = False

#########
# get job_template list 
#########

#get files under folder job-templates
job_templates_files = os.listdir("job-templates")
#remove file name ext: .job
job_templates = [e[:-4] for e in job_templates_files]

#########
# validate bids-apps.tsv
#########

line_index = 0
with open("bids-apps.tsv") as fd:
    rd = csv.reader(fd, delimiter="\t", quotechar='"')
    for row in rd:

        # line# start with 1
        line_index = line_index+1

        #each row must has 4 columns
        if len(row) != 4:
            raise ValueError("LINE {} ERROR: each row must has 4 columns".format(line_index))

        #parse
        name=row[0]
        url=row[1]
        base_options=row[2]
        job_template=row[3]

        # skip 1st row
        if name=="name":
            continue
        
        # skip comment
        if name[0]=="#":
            continue
        
        #########
        #url must star with "docker://" or "shub://"
        #########

        if url.startswith("docker://") or url.startswith("shub://") or url.startswith("file://"):
            pass
        else:
            raise ValueError("LINE {} ERROR: url must star with 'docker://' or 'shub://'  or 'file://' ".format(line_index))    
        
        #########
        #docker_image_name_and_tag == name
        #    docker://khanlab/gradcorrect:v0.0.2 docker_image_name_and_tag： gradcorrect:v0.0.2
        #    name: gradcorrect_v0.0.2 
        
        # or 

        #docker_image_name_and_tag in name 
        #    docker://khanlab/prepdwi:v0.0.12
        #    prepdwi_v0.0.12d_bedpost
        #########

        # docker_image_name_and_tag = url.split('/')[-1]
        # replaced = docker_image_name_and_tag.replace(":","_")
        # if  replaced not in name:
        #     raise ValueError("LINE {} ERROR: name NOT match with docker image's name and tag".format(line_index))    
        
        #########
        # check if docker/singularity image exist
        #########
        if url.startswith("docker://"):
            # remove images after pull to avoid "no space left on device"
            cmd = "docker pull {} > /dev/null && docker rmi -f $(docker images -q)".format(url[9:])
            print(cmd)
            if validate_image:
                os.system(cmd)

        if url.startswith("shub://"):
            cmd = "singularity pull {} --name temp.img  > /dev/null && rm temp.img".format(url[7:])
            print(cmd)
            if validate_image:
                os.system(cmd)

        #########
        # base_options must be '' or start with '-'
        #########
        if base_options =="" or base_options.startswith(""):
            pass
        else:
            raise ValueError("LINE {} ERROR: base_options must be '' or start with '-' ".format(line_index))    
       
        #########
        #job_template must be a string match a job name in folder job-templates
        #########
        if job_template not in job_templates:
            raise ValueError("LINE {} ERROR: job_template must match with one job in job-templates".format(line_index))    

        


from __future__ import absolute_import, annotations

import os
from pathlib import Path
from typing import NoReturn

import appdirs

from kslurm.utils import get_hash

CACHE_PATH = Path(appdirs.user_cache_dir("kslurm"))


class Cache:
    def __init__(self):
        self._path = CACHE_PATH
        self._path.mkdir(exist_ok=True)

    def get_path(self, key: str):
        return self._path / get_hash(key)

    def __getitem__(self, key: str):
        path = self.get_path(key)
        if not path.exists():
            raise KeyError(f"{key} not found")
        with path.open("r") as f:
            return f.read()

    def __setitem__(self, key: str, data: str):
        path = self.get_path(key)
        with path.open("w") as f:
            f.write(data)

    def __delitem__(self, key: str):
        path = self.get_path(key)
        if not path.exists():
            raise KeyError(f"{key} not found")
        os.remove(path)

    def __len__(self):
        return len(list(self._path.iterdir()))

    def __iter__(self) -> NoReturn:
        raise NotImplementedError()
